import os
import warnings

import gym
import pytest

from plangym.control.classic_control import ClassicControl
from plangym.environment_names import ATARI, BOX_2D, CLASSIC_CONTROL, DM_CONTROL, RETRO
from plangym.registry import make
from plangym.vectorization.parallel import ParallelEnv
from tests import (
    SKIP_ATARI_TESTS,
    SKIP_BOX2D_TESTS,
    SKIP_DM_CONTROL_TESTS,
    SKIP_RAY_TESTS,
    SKIP_RETRO_TESTS,
)


def _test_env_class(name, cls, **kwargs):
    n_workers = 2
    assert isinstance(make(name, delay_setup=False, **kwargs), cls)
    env = make(name=name, n_workers=n_workers, delay_setup=True, **kwargs)
    assert isinstance(env, ParallelEnv)
    assert env._env_class == cls
    assert env.n_workers == n_workers
    if not SKIP_RAY_TESTS:
        from plangym.vectorization.ray import RayEnv

        env = make(name=name, n_workers=n_workers, ray=True, delay_setup=True, **kwargs)
        assert isinstance(env, RayEnv)
        assert env._env_class == cls
        assert env.n_workers == n_workers


class TestMake:
    @pytest.mark.parametrize("name", CLASSIC_CONTROL)
    def test_classic_control_make(self, name):
        _test_env_class(name, ClassicControl)

    @pytest.mark.skipif(SKIP_ATARI_TESTS, reason="Atari not installed")
    @pytest.mark.parametrize("name", ATARI[::10])
    def test_atari_make(self, name):
        from plangym.videogames.atari import AtariEnv

        _test_env_class(name, AtariEnv)

    @pytest.mark.skipif(SKIP_BOX2D_TESTS, reason="BOX_2D not installed")
    @pytest.mark.parametrize("name", BOX_2D)
    def test_box2d_make(self, name):
        from plangym.control.box_2d import Box2DEnv
        from plangym.control.lunar_lander import LunarLander

        if name == "FastLunarLander-v0":
            _test_env_class(name, LunarLander)
            return
        elif name == "CarRacing-v0" and os.getenv("SKIP_RENDER", False):
            return
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            _test_env_class(name, Box2DEnv)

    @pytest.mark.skipif(SKIP_RETRO_TESTS, reason="Retro not installed")
    @pytest.mark.parametrize("name", RETRO[::10])
    def test_retro_make(self, name):
        from plangym.videogames.retro import RetroEnv

        try:
            _test_env_class(name, RetroEnv)
        except FileNotFoundError:
            pass

    @pytest.mark.skipif(SKIP_RETRO_TESTS, reason="Retro not installed")
    def test_retro_make_with_state(self):
        from plangym.videogames.retro import ActionDiscretizer, RetroEnv

        try:
            _test_env_class(
                "SonicTheHedgehog-Genesis",
                RetroEnv,
                state="GreenHillZone.Act3",
                wrappers=[ActionDiscretizer],
            )
        except FileNotFoundError:
            pass

    @pytest.mark.skipif(SKIP_ATARI_TESTS, reason="Atari not installed")
    def test_custom_atari_make(self):
        # from plangym.minimal import MinimalPacman, MinimalPong
        from plangym.videogames import MontezumaEnv

        # _test_env_class("MinimalPacman-v0", MinimalPacman)
        #  _test_env_class("MinimalPong-v0", MinimalPong)
        _test_env_class("PlanMontezuma-v0", MontezumaEnv)

    @pytest.mark.skipif(SKIP_DM_CONTROL_TESTS, reason="dm_control not installed")
    @pytest.mark.parametrize("name", DM_CONTROL)
    def test_dmcontrol_make(self, name):
        from plangym.control.dm_control import DMControlEnv

        domain_name, task_name = name
        if task_name is not None:
            _test_env_class(domain_name, DMControlEnv, task_name=task_name)
        else:
            _test_env_class(domain_name, DMControlEnv)

    @pytest.mark.skipif(SKIP_DM_CONTROL_TESTS, reason="dm_control not installed")
    @pytest.mark.parametrize("name", DM_CONTROL)
    def test_dmcontrol_domain_name_make(self, name):
        from plangym.control.dm_control import DMControlEnv

        domain_name, task_name = name
        if task_name is not None:
            _test_env_class(
                name=None, domain_name=domain_name, cls=DMControlEnv, task_name=task_name
            )
        else:
            _test_env_class(name=None, domain_name=domain_name, cls=DMControlEnv)

    def test_invalid_name(self):
        with pytest.raises(gym.error.Error):
            make(name="Miaudb")
        with pytest.raises(gym.error.UnregisteredEnv):
            make(name="Miaudb-v0")



"""Plangym API implementation."""
from abc import ABC
from typing import Any, Dict, Iterable, Optional

import gym
import numpy

from plangym.core import PlangymEnv, wrap_callable


LIFE_KEY = "lifes"


class VideogameEnv(PlangymEnv, ABC):
    """Common interface for working with video games that run using an emulator."""

    AVAILABLE_OBS_TYPES = {"rgb", "grayscale", "ram"}
    DEFAULT_OBS_TYPE = "rgb"

    def __init__(
        self,
        name: str,
        frameskip: int = 5,
        episodic_life: bool = False,
        autoreset: bool = True,
        delay_setup: bool = False,
        remove_time_limit: bool = True,
        obs_type: str = "rgb",  # ram | rgb | grayscale
        render_mode: Optional[str] = None,  # None | human | rgb_array
        wrappers: Iterable[wrap_callable] = None,
        **kwargs,
    ):
        """
        Initialize a :class:`VideogameEnv`.

        Args:
            name: Name of the environment. Follows standard gym syntax conventions.
            frameskip: Number of times an action will be applied for each step
                in dt.
            episodic_life: Return ``end = True`` when losing a life.
            autoreset: Restart environment when reaching a terminal state.
            delay_setup: If ``True`` do not initialize the ``gym.Environment``
                and wait for ``setup`` to be called later.
            remove_time_limit: If True, remove the time limit from the environment.
            obs_type: One of {"rgb", "ram", "grayscale"}.
            mode: Integer or string indicating the game mode, when available.
            difficulty: Difficulty level of the game, when available.
            repeat_action_probability: Repeat the last action with this probability.
            full_action_space: Whether to use the full range of possible actions
                               or only those available in the game.
            render_mode: One of {None, "human", "rgb_aray"}.
            wrappers: Wrappers that will be applied to the underlying OpenAI env.
                      Every element of the iterable can be either a :class:`gym.Wrapper`
                      or a tuple containing ``(gym.Wrapper, kwargs)``.

        """
        self.episodic_life = episodic_life
        self._info_step = {LIFE_KEY: -1, "lost_life": False}
        super(VideogameEnv, self).__init__(
            name=name,
            frameskip=frameskip,
            autoreset=autoreset,
            wrappers=wrappers,
            delay_setup=delay_setup,
            render_mode=render_mode,
            remove_time_limit=remove_time_limit,
            obs_type=obs_type,
            **kwargs,
        )

    @property
    def n_actions(self) -> int:
        """Return the number of actions available."""
        return self.action_space.n

    @staticmethod
    def get_lifes_from_info(info: Dict[str, Any]) -> int:
        """Return the number of lifes remaining in the current game."""
        return info.get("life", -1)

    def apply_action(self, action):
        """Evolve the environment for one time step applying the provided action."""
        obs, reward, terminal, info = super(VideogameEnv, self).apply_action(action=action)
        info[LIFE_KEY] = self.get_lifes_from_info(info)
        past_lifes = self._info_step.get(LIFE_KEY, -1)
        lost_life = past_lifes > info[LIFE_KEY] or self._info_step.get("lost_life")
        info["lost_life"] = lost_life
        terminal = (terminal or lost_life) if self.episodic_life else terminal
        return obs, reward, terminal, info

    def clone(self, **kwargs) -> "VideogameEnv":
        """Return a copy of the environment."""
        params = dict(
            episodic_life=self.episodic_life,
            obs_type=self.obs_type,
            render_mode=self.render_mode,
        )
        params.update(**kwargs)
        return super(VideogameEnv, self).clone(**params)

    def begin_step(self, action=None, dt=None, state=None, return_state: bool = None) -> None:
        """Perform setup of step variables before starting `step_with_dt`."""
        self._info_step = {LIFE_KEY: -1, "lost_life": False}
        super(VideogameEnv, self).begin_step(
            action=action,
            dt=dt,
            state=state,
            return_state=return_state,
        )

    def init_spaces(self) -> None:
        """Initialize the action_space and the observation_space of the environment."""
        super(VideogameEnv, self).init_spaces()
        if self.obs_type == "ram":
            if self.DEFAULT_OBS_TYPE == "ram":
                space = self.gym_env.observation_space
            else:
                ram_size = self.get_ram().shape
                space = gym.spaces.Box(low=0, high=255, dtype=numpy.uint8, shape=ram_size)
            self._obs_space = space

    def process_obs(self, obs, **kwargs):
        """Return the ram vector if obs_type == "ram" or and image otherwise."""
        obs = super(VideogameEnv, self).process_obs(obs, **kwargs)
        if self.obs_type == "ram" and self.DEFAULT_OBS_TYPE != "ram":
            obs = self.get_ram()
        return obs

    def get_ram(self) -> numpy.ndarray:
        """Return the ram of the emulator as a numpy array."""
        raise NotImplementedError()



import warnings

import pytest


warnings.filterwarnings(
    action="ignore", category=pytest.PytestUnraisableExceptionWarning, module="pytest"
)
try:
    import retro

    SKIP_RETRO_TESTS = False
except ImportError:
    SKIP_RETRO_TESTS = True

try:
    import ray

    SKIP_RAY_TESTS = False
except ImportError:
    SKIP_RAY_TESTS = True

try:
    from plangym.videogames.atari import AtariEnv

    SKIP_ATARI_TESTS = False
except ImportError:
    SKIP_ATARI_TESTS = True

try:
    from plangym.control.dm_control import DMControlEnv

    DMControlEnv(name="walker-run", frameskip=3)
    SKIP_DM_CONTROL_TESTS = False
except ImportError:
    SKIP_DM_CONTROL_TESTS = True


try:
    import Box2D

    SKIP_BOX2D_TESTS = False
except ImportError:
    SKIP_BOX2D_TESTS = True



from typing import Union

import gym
import pytest

from plangym.vectorization.parallel import ParallelEnv
from plangym.videogames.retro import ActionDiscretizer, RetroEnv


pytest.importorskip("retro")
from plangym.api_tests import batch_size, display, TestPlanEnv, TestPlangymEnv  # noqa: F401


def retro_airstrike():
    res_obs = gym.wrappers.resize_observation.ResizeObservation
    return RetroEnv(name="Airstriker-Genesis", wrappers=[(res_obs, {"shape": (90, 90)})])


def retro_sonic():

    return RetroEnv(
        name="SonicTheHedgehog-Genesis",
        state="GreenHillZone.Act3",
        wrappers=[ActionDiscretizer],
        obs_type="grayscale",
    )


def parallel_retro():
    return ParallelEnv(
        name="Airstriker-Genesis",
        env_class=RetroEnv,
        n_workers=2,
        obs_type="ram",
        wrappers=[ActionDiscretizer],
    )


environments = [retro_airstrike, retro_sonic, parallel_retro]


@pytest.fixture(params=environments, scope="class")
def env(request) -> Union[RetroEnv, ParallelEnv]:
    env_ = request.param()
    if env_.delay_setup and env_.gym_env is None:
        env_.setup()
    yield env_
    env_.close()


class TestRetro:
    def test_init_env(self):
        env = retro_airstrike()
        env.reset()
        env.setup()

    def test_getattribute(self):
        env = retro_airstrike()
        env.em.get_state()

    def test_clone(self):
        env = RetroEnv(name="Airstriker-Genesis", obs_type="ram", delay_setup=True)
        new_env = env.clone()
        del env
        new_env.reset()



import os

import numpy
import pytest


pytest.importorskip("dm_control")
from plangym.api_tests import (  # noqa: F401
    batch_size,
    display,
    generate_test_cases,
    TestPlanEnv,
    TestPlangymEnv,
)
from plangym.control.dm_control import DMControlEnv
from plangym.environment_names import DM_CONTROL


class DummyTimeLimit:
    def __init__(self, env, max_episode_steps=None):
        self._max_episode_steps = max_episode_steps
        self._elapsed_steps = None
        self.env = env

    def __getattr__(self, item):
        return getattr(self.env, item)


def walker_run():
    timelimit = [(DummyTimeLimit, {"max_episode_steps": 1000})]
    return DMControlEnv(name="walker-run", frameskip=3, wrappers=timelimit)


def parallel_dm():
    return DMControlEnv(name="cartpole-balance", frameskip=3)


environments = [walker_run, parallel_dm]


@pytest.fixture(
    params=generate_test_cases(DM_CONTROL, DMControlEnv, n_workers_values=[None]),
    scope="module",
)
def env(request) -> DMControlEnv:
    env = request.param()
    yield env
    try:
        env.close()
    except Exception:
        pass


class TestDMControl:
    def test_attributes(self, env):
        env.reset()
        assert hasattr(env, "physics")
        assert hasattr(env, "action_spec")
        assert hasattr(env, "action_space")
        assert hasattr(env, "render_mode")
        assert env.render_mode in {"human", "rgb_array", "coords", None}

    @pytest.mark.skipif(os.getenv("SKIP_RENDER", False), reason="No display in CI.")
    def test_render(self, env):
        env.reset()
        obs_rgb = env.render(mode="rgb_array")
        assert isinstance(obs_rgb, numpy.ndarray)
        old_len = len(env.viewer)
        action = env.sample_action()
        env.step(action)
        env.render(mode="human")
        assert len(env.viewer) > old_len
        env.show_game(sleep=0.01)

    def test_parse_name_fails(self):
        with pytest.raises(ValueError):
            DMControlEnv(name="cartpole")



"""Implement the ``plangym`` API for ``dm_control`` environments."""
from typing import Iterable, Optional
import warnings

from gym.spaces import Box
import numpy

from plangym.core import PlangymEnv, wrap_callable


try:
    from gym.envs.classic_control import rendering

    novideo_mode = False
except Exception:  # pragma: no cover
    novideo_mode = True


class DMControlEnv(PlangymEnv):
    """
    Wrap the `dm_control library, allowing its implementation in planning problems.

    The dm_control library is a DeepMind's software stack for physics-based
    simulation and Reinforcement Learning environments, using MuJoCo physics.

    For more information about the environment, please refer to
    https://github.com/deepmind/dm_control

    This class allows the implementation of dm_control in planning problems.
    It allows parallel and vectorized execution of the environments.
    """

    DEFAULT_OBS_TYPE = "coords"

    def __init__(
        self,
        name: str = "cartpole-balance",
        frameskip: int = 1,
        episodic_life: bool = False,
        autoreset: bool = True,
        wrappers: Iterable[wrap_callable] = None,
        delay_setup: bool = False,
        visualize_reward: bool = True,
        domain_name=None,
        task_name=None,
        render_mode=None,
        obs_type: Optional[str] = None,
        remove_time_limit=None,
    ):
        """
        Initialize a :class:`DMControlEnv`.

        Args:
            name: Name of the task. Provide the task to be solved as
                `domain_name-task_name`. For example 'cartpole-balance'.
            frameskip: Set a deterministic frameskip to apply the same
                action N times.
            episodic_life: Send terminal signal after loosing a life.
            autoreset: Restart environment when reaching a terminal state.
            wrappers: Wrappers that will be applied to the underlying OpenAI env. \
                Every element of the iterable can be either a :class:`gym.Wrapper` \
                or a tuple containing ``(gym.Wrapper, kwargs)``.
            delay_setup: If ``True``, do not initialize the ``gym.Environment`` \
                and wait for ``setup`` to be called later.
            visualize_reward: Define the color of the agent, which depends
                on the reward on its last timestep.
            domain_name: Same as in dm_control.suite.load.
            task_name: Same as in dm_control.suite.load.
            render_mode: None|human|rgb_array
        """
        self._visualize_reward = visualize_reward
        self.viewer = []
        self._viewer = None
        name, self._domain_name, self._task_name = self._parse_names(name, domain_name, task_name)
        super(DMControlEnv, self).__init__(
            name=name,
            frameskip=frameskip,
            episodic_life=episodic_life,
            wrappers=wrappers,
            delay_setup=delay_setup,
            autoreset=autoreset,
            render_mode=render_mode,
            obs_type=obs_type,
        )

    @property
    def physics(self):
        """Alias for gym_env.physics."""
        return self.gym_env.physics

    @property
    def domain_name(self) -> str:
        """Return the name of the agent in the current simulation."""
        return self._domain_name

    @property
    def task_name(self) -> str:
        """Return the name of the task in the current simulation."""
        return self._task_name

    @staticmethod
    def _parse_names(name, domain_name, task_name):
        """Return the name, domain name, and task name of the project."""
        if isinstance(name, str) and domain_name is None:
            domain_name = name if "-" not in name else name.split("-")[0]

        if isinstance(name, str) and "-" in name and task_name is None:
            task_name = task_name if "-" not in name else name.split("-")[1]
        if (not isinstance(name, str) or "-" not in name) and task_name is None:
            raise ValueError(
                f"Invalid combination: name {name},"
                f" domain_name {domain_name}, task_name {task_name}",
            )
        name = "-".join([domain_name, task_name])
        return name, domain_name, task_name

    def init_gym_env(self):
        """Initialize the environment instance (dm_control) that the current class is wrapping."""
        from dm_control import suite

        env = suite.load(
            domain_name=self.domain_name,
            task_name=self.task_name,
            visualize_reward=self._visualize_reward,
        )
        self.viewer = []
        self._viewer = None if novideo_mode else rendering.SimpleImageViewer()
        return env

    def setup(self):
        """Initialize the target :class:`gym.Env` instance."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            super(DMControlEnv, self).setup()

    def _init_action_space(self):
        """
        Define the action space of the environment.

        This method determines the spectrum of possible actions that the
        agent can perform. The action space consists in a grid representing
        the Cartesian product of the closed intervals defined by the user.
        """
        self._action_space = Box(
            low=self.action_spec().minimum,
            high=self.action_spec().maximum,
            dtype=numpy.float32,
        )

    def _init_obs_space_coords(self):
        """Define the observation space of the environment."""
        shape = self.reset(return_state=False).shape
        self._obs_space = Box(low=-numpy.inf, high=numpy.inf, shape=shape, dtype=numpy.float32)

    def action_spec(self):
        """Alias for the environment's ``action_spec``."""
        return self.gym_env.action_spec()

    def get_image(self) -> numpy.ndarray:
        """
        Return a numpy array containing the rendered view of the environment.

        Square matrices are interpreted as a greyscale image. Three-dimensional arrays
        are interpreted as RGB images with channels (Height, Width, RGB).
        """
        return self.gym_env.physics.render(camera_id=0)

    def render(self, mode="human"):
        """
        Store all the RGB images rendered to be shown when the `show_game`\
        function is called.

        Args:
            mode: `rgb_array` return an RGB image stored in a numpy array. `human`
             stores the rendered image in a viewer to be shown when `show_game`
             is called.

        Returns:
            numpy.ndarray when mode == `rgb_array`. True when mode == `human`
        """
        img = self.get_image()
        if mode == "rgb_array":
            return img
        elif mode == "human":
            self.viewer.append(img)
        return True

    def show_game(self, sleep: float = 0.05):
        """
        Render the collected RGB images.

        When 'human' option is selected as argument for the `render` method,
        it stores a collection of RGB images inside the ``self.viewer``
        attribute. This method calls the latter to visualize the collected
        images.
        """
        import time

        for img in self.viewer:
            self._viewer.imshow(img)
            time.sleep(sleep)

    def get_coords_obs(self, obs, **kwargs) -> numpy.ndarray:
        """
        Get the environment observation from a time_step object.

        Args:
            obs: Time step object returned after stepping the environment.
            **kwargs: Ignored

        Returns:
            Numpy array containing the environment observation.
        """
        return self._time_step_to_obs(time_step=obs)

    def set_state(self, state: numpy.ndarray) -> None:
        """
        Set the state of the simulator to the target State.

        Args:
            state: numpy.ndarray containing the information about the state to be set.

        Returns:
            None
        """
        with self.gym_env.physics.reset_context():
            self.gym_env.physics.set_state(state)

    def get_state(self) -> numpy.ndarray:
        """
        Return a tuple containing the three arrays that characterize the state\
         of the system.

        Each tuple contains the position of the robot, its velocity
         and the control variables currently being applied.

        Returns:
            Tuple of numpy arrays containing all the information needed to describe
            the current state of the simulation.
        """
        return self.gym_env.physics.get_state()

    def apply_action(self, action):
        """Transform the returned time_step object to a compatible gym tuple."""
        info = {}
        time_step = self.gym_env.step(action)
        obs = time_step
        terminal = time_step.last()
        _reward = time_step.reward if time_step.reward is not None else 0.0
        reward = _reward + self._reward_step
        return obs, reward, terminal, info

    @staticmethod
    def _time_step_to_obs(time_step) -> numpy.ndarray:
        """
        Stack observation values as a horizontal sequence.

        Concat observations in a single array, making easier calculating
        distances.
        """
        obs_array = numpy.hstack(
            [numpy.array([time_step.observation[x]]).flatten() for x in time_step.observation],
        )
        return obs_array

    def close(self):
        """Tear down the environment and close rendering."""
        try:
            super(DMControlEnv, self).close()
            if self._viewer is not None:
                self._viewer.close()
        except Exception:
            pass



"""Module that contains environments representing control tasks."""
from plangym.control.balloon import BalloonEnv  # noqa: E402
from plangym.control.box_2d import Box2DEnv  # noqa: E402
from plangym.control.classic_control import ClassicControl  # noqa: E402
from plangym.control.dm_control import DMControlEnv  # noqa: E402
from plangym.control.lunar_lander import LunarLander  # noqa: E402



import os

import pytest


pytest.importorskip("balloon_learning_environment")
from plangym.api_tests import (  # noqa: F401
    batch_size,
    display,
    generate_test_cases,
    TestPlanEnv,
    TestPlangymEnv,
)
from plangym.control.balloon import BalloonEnv


disable_balloon_tests = os.getenv("DISABLE_BALLOON_ENV", True)
if disable_balloon_tests and str(disable_balloon_tests).lower() != "false":
    pytest.skip("balloon_learning_environment tests are disabled", allow_module_level=True)


@pytest.fixture(
    params=generate_test_cases(["BalloonLearningEnvironment-v0"], BalloonEnv),
    scope="module",
)
def env(request) -> BalloonEnv:
    return request.param()



"""Implementation of the montezuma environment adapted for planning problems."""
import pickle
from typing import Iterable, Optional, Tuple, Union

import cv2
import gym
from gym.envs.registration import registry as gym_registry
import numpy

from plangym.core import wrap_callable
from plangym.videogames.atari import AtariEnv


# ------------------------------------------------------------------------------
# Copyright (c) 2018-2019 Uber Technologies, Inc.
#
# Licensed under the Uber Non-Commercial License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at the root directory of this project.
#
# See the License for the specific language governing permissions and
# limitations under the License.


class MontezumaPosLevel:
    """Contains the information of Panama Joe."""

    __slots__ = ["level", "score", "room", "x", "y", "tuple"]

    def __init__(self, level, score, room, x, y):
        """Initialize a :class:`MontezumaPosLevel`."""
        self.level = level
        self.score = score
        self.room = room
        self.x = x
        self.y = y
        self.tuple = None
        self.set_tuple()

    def set_tuple(self):
        """Set the tuple values from the other attributes."""
        self.tuple = (self.level, self.score, self.room, self.x, self.y)

    def __hash__(self):
        """Hash the current instance."""
        return hash(self.tuple)

    def __eq__(self, other):
        """Compare equality between instances."""
        if not isinstance(other, MontezumaPosLevel):
            return False
        return self.tuple == other.tuple

    def __getstate__(self):
        """Return the tuple containing the attributes."""
        return self.tuple

    def __setstate__(self, d):
        """Set the class attributes using the provided tuple."""
        self.level, self.score, self.room, self.x, self.y = d
        self.tuple = d

    def __repr__(self):
        """Print the attributes of the current instance."""
        return f"Level={self.level} Room={self.room} Objects={self.score} x={self.x} y={self.y}"


PYRAMID = [
    [-1, -1, -1, 0, 1, 2, -1, -1, -1],
    [-1, -1, 3, 4, 5, 6, 7, -1, -1],
    [-1, 8, 9, 10, 11, 12, 13, 14, -1],
    [15, 16, 17, 18, 19, 20, 21, 22, 23],
]

OBJECT_PIXELS = [
    50,  # Hammer/mallet
    40,  # Key 1
    40,  # Key 2
    40,  # Key 3
    37,  # Sword 1
    37,  # Sword 2
    42,  # Torch
]

KNOWN_XY = [None] * 24

KEY_BITS = 0x8 | 0x4 | 0x2


class CustomMontezuma:
    """MontezumaEnv environment that tracks the room and position of Panama Joe."""

    def __init__(
        self,
        check_death: bool = True,
        obs_type: str = "rgb",
        score_objects: bool = False,
        objects_from_pixels: bool = False,
        objects_remember_rooms: bool = False,
        only_keys: bool = False,
        death_room_8: bool = True,
    ):  # TODO: version that also considers the room objects were found in
        """Initialize a :class:`CustomMontezuma`."""
        spec = gym_registry.spec("MontezumaRevengeDeterministic-v4")
        # not actually needed, but we feel safer
        spec.max_episode_steps = int(1e100)
        spec.max_episode_time = int(1e100)
        self.env = spec.make()
        self.env.reset()
        self.score_objects = score_objects
        self.ram = None
        self.check_death = check_death
        self.cur_steps = 0
        self.cur_score = 0
        self.rooms = {}
        self.room_time = (None, None)
        self.room_threshold = 40
        self.unwrapped.seed(0)
        self.coords_obs = obs_type == "coords"
        self.state = []
        self.ram_death_state = -1
        self._x_repeat = 2
        self._death_room_8 = death_room_8
        self.cur_lives = 5
        self.ignore_ram_death = False
        self.objects_from_pixels = objects_from_pixels
        self.objects_remember_rooms = objects_remember_rooms
        self.only_keys = only_keys
        self.pos = MontezumaPosLevel(0, 0, 0, 0, 0)
        if self.coords_obs:
            shape = self.get_coords().shape
            self.observation_space = gym.spaces.Box(
                low=-numpy.inf,
                high=numpy.inf,
                dtype=numpy.float32,
                shape=shape,
            )

    def __getattr__(self, e):
        """Forward to gym environment."""
        return getattr(self.env, e)

    def reset(self, seed=None, return_info: bool = False) -> numpy.ndarray:
        """Reset the environment."""
        obs = self.env.reset()
        self.cur_lives = 5
        for _ in range(3):
            obs, *_, info = self.env.step(0)
        self.ram = self.env.unwrapped.ale.getRAM()
        self.cur_score = 0
        self.cur_steps = 0
        self.ram_death_state = -1
        self.pos = None
        self.pos = self.pos_from_obs(
            self.get_face_pixels(obs),
            obs,
        )
        if self.get_pos().room not in self.rooms:
            self.rooms[self.get_pos().room] = (
                False,
                obs[50:].repeat(self._x_repeat, axis=1),
            )
        self.room_time = (self.get_pos().room, 0)
        if self.coords_obs:
            return self.get_coords()
        return obs

    def step(self, action) -> Tuple[numpy.ndarray, float, bool, dict]:
        """Step the environment."""
        obs, reward, done, info = self.env.step(action)
        self.ram = self.env.unwrapped.ale.getRAM()
        self.cur_steps += 1

        face_pixels = self.get_face_pixels(obs)
        pixel_death = self.is_pixel_death(obs, face_pixels)
        ram_death = self.is_ram_death()
        # TODO: remove all this stuff
        if self.check_death and pixel_death:
            done = True
        elif self.check_death and not pixel_death and ram_death:
            done = True

        self.cur_score += reward
        self.pos = self.pos_from_obs(face_pixels, obs)
        if self.pos.room != self.room_time[0]:  # pragma: no cover
            self.room_time = (self.pos.room, 0)
        self.room_time = (self.pos.room, self.room_time[1] + 1)
        if self.pos.room not in self.rooms or (
            self.room_time[1] == self.room_threshold and not self.rooms[self.pos.room][0]
        ):
            self.rooms[self.pos.room] = (
                self.room_time[1] == self.room_threshold,
                obs[50:].repeat(self._x_repeat, axis=1),
            )
        if self._death_room_8:
            done = done or self.pos.room == 8
        if self.coords_obs:
            return self.get_coords(), reward, done, info
        return obs, reward, done, info

    def pos_from_obs(self, face_pixels, obs) -> MontezumaPosLevel:
        """Extract the information of the position of Panama Joe."""
        face_pixels = [(y, x * self._x_repeat) for y, x in face_pixels]
        if len(face_pixels) == 0:
            assert self.pos is not None, "No face pixel and no previous pos"
            return self.pos  # Simply re-use the same position
        y, x = numpy.mean(face_pixels, axis=0)
        room = 1
        level = 0
        old_objects = tuple()
        if self.pos is not None:
            room = self.pos.room
            level = self.pos.level
            old_objects = self.pos.score
            direction_x = numpy.clip(int((self.pos.x - x) / 50), -1, 1)
            direction_y = numpy.clip(int((self.pos.y - y) / 50), -1, 1)
            if direction_x != 0 or direction_y != 0:  # pragma: no cover
                room_x, room_y = self.get_room_xy(self.pos.room)
                if room == 15 and room_y + direction_y >= len(PYRAMID):
                    room = 1
                    level += 1
                else:
                    assert (
                        direction_x == 0 or direction_y == 0
                    ), f"Room change in more than two directions : ({direction_y}, {direction_x})"
                    room = PYRAMID[room_y + direction_y][room_x + direction_x]
                    assert room != -1, f"Impossible room change: ({direction_y}, {direction_x})"

        score = self.cur_score
        if self.score_objects:  # TODO: detect objects from the frame!
            if not self.objects_from_pixels:
                score = self.ram[65]
                if self.only_keys:  # pragma: no cover
                    # These are the key bytes
                    score &= KEY_BITS
            else:
                score = self.get_objects_from_pixels(obs, room, old_objects)
        return MontezumaPosLevel(level, score, room, x, y)

    def get_objects_from_pixels(self, obs, room, old_objects):
        """Extract the position of the objects in the provided observation."""
        object_part = (obs[25:45, 55:110, 0] != 0).astype(numpy.uint8) * 255
        connected_components = cv2.connectedComponentsWithStats(object_part)
        pixel_areas = list(e[-1] for e in connected_components[2])[1:]

        if self.objects_remember_rooms:
            cur_object = []
            old_objects = list(old_objects)
            for _, n_pixels in enumerate(OBJECT_PIXELS):
                if n_pixels != 40 and self.only_keys:  # pragma: no cover
                    continue
                if n_pixels in pixel_areas:  # pragma: no cover
                    pixel_areas.remove(n_pixels)
                    orig_types = [e[0] for e in old_objects]
                    if n_pixels in orig_types:
                        idx = orig_types.index(n_pixels)
                        cur_object.append((n_pixels, old_objects[idx][1]))
                        old_objects.pop(idx)
                    else:
                        cur_object.append((n_pixels, room))

            return tuple(cur_object)

        else:
            cur_object = 0
            for i, n_pixels in enumerate(OBJECT_PIXELS):
                if n_pixels in pixel_areas:  # pragma: no cover
                    pixel_areas.remove(n_pixels)
                    cur_object |= 1 << i

            if self.only_keys:
                # These are the key bytes
                cur_object &= KEY_BITS
            return cur_object

    def get_coords(self) -> numpy.ndarray:
        """Return an observation containing the position and the flattened screen of the game."""
        coords = numpy.array([self.pos.x, self.pos.y, self.pos.room, self.score_objects])
        return coords

    def state_to_numpy(self) -> numpy.ndarray:
        """Return a numpy array containing the current state of the game."""
        state = self.unwrapped.clone_state(include_rng=False)
        state = numpy.frombuffer(pickle.dumps(state), dtype=numpy.uint8)
        return state

    def _restore_state(self, state) -> None:
        """Restore the state of the game from the provided numpy array."""
        state = pickle.loads(state.tobytes())
        self.unwrapped.restore_state(state)

    def get_restore(self) -> tuple:
        """Return a tuple containing all the information needed to clone the state of the env."""
        return (
            self.state_to_numpy(),
            self.cur_score,
            self.cur_steps,
            self.pos,
            self.room_time,
            self.ram_death_state,
            self.score_objects,
            self.cur_lives,
        )

    def restore(self, data) -> None:
        """Restore the state of the env from the provided tuple."""
        (
            full_state,
            score,
            steps,
            pos,
            room_time,
            ram_death_state,
            self.score_objects,
            self.cur_lives,
        ) = data
        self.env.reset()
        self._restore_state(full_state)
        self.ram = self.env.unwrapped.ale.getRAM()
        self.cur_score = score
        self.cur_steps = steps
        self.pos = pos
        self.room_time = room_time
        assert len(self.room_time) == 2
        self.ram_death_state = ram_death_state

    def is_transition_screen(self, obs) -> bool:
        """Return True if the current observation corresponds to a transition between rooms."""
        obs = obs[50:, :, :]
        # The screen is a transition screen if it is all black or if its color is made
        # up only of black and (0, 28, 136), which is a color seen in the transition
        # screens between two levels.
        unprocessed_one = obs[:, :, 1]
        unprocessed_two = obs[:, :, 2]
        return (
            numpy.sum(obs[:, :, 0] == 0)
            + numpy.sum((unprocessed_one == 0) | (unprocessed_one == 28))
            + numpy.sum((unprocessed_two == 0) | (unprocessed_two == 136))
        ) == obs.size

    def get_face_pixels(self, obs) -> set:
        """Return the pixels containing the face of Paname Joe."""
        # TODO: double check that this color does not re-occur somewhere else
        # in the environment.
        return set(zip(*numpy.where(obs[50:, :, 0] == 228)))

    def is_pixel_death(self, obs, face_pixels):
        """Return a death signal extracted from the observation of the environment."""
        # There are no face pixels and yet we are not in a transition screen. We
        # must be dead!
        if len(face_pixels) == 0:
            # All of the screen except the bottom is black: this is not a death but a
            # room transition. Ignore.
            if self.is_transition_screen(obs):  # pragma: no cover
                return False
            return True

        # We already checked for the presence of no face pixels, however,
        # sometimes we can die and still have face pixels. In those cases,
        # the face pixels will be DISCONNECTED.
        for pixel in face_pixels:
            for neighbor in [(-1, 0), (1, 0), (0, -1), (0, 1)]:
                if (
                    pixel[0] + neighbor[0],
                    pixel[1] + neighbor[1],
                ) in face_pixels:  # pragma: no cover
                    return False

        return True  # pragma: no cover

    def is_ram_death(self) -> bool:
        """Return a death signal extracted from the ram of the environment."""
        if self.ram[58] > self.cur_lives:  # pragma: no cover
            self.cur_lives = self.ram[58]
        return self.ram[55] != 0 or self.ram[58] < self.cur_lives

    def get_pos(self) -> MontezumaPosLevel:
        """Return the current pos."""
        assert self.pos is not None
        return self.pos

    @staticmethod
    def get_room_xy(room) -> Tuple[Union[None, Tuple[int, int]]]:
        """Get the tuple that encodes the provided room."""
        if KNOWN_XY[room] is None:
            for y, l in enumerate(PYRAMID):
                if room in l:
                    KNOWN_XY[room] = (l.index(room), y)
                    break
        return KNOWN_XY[room]

    @staticmethod
    def get_room_out_of_bounds(room_x, room_y) -> bool:
        """Return a boolean indicating if the provided tuple represents and invalid room."""
        return room_y < 0 or room_x < 0 or room_y >= len(PYRAMID) or room_x >= len(PYRAMID[0])

    @staticmethod
    def get_room_from_xy(room_x, room_y) -> int:
        """Return the number of the room from a tuple."""
        return PYRAMID[room_y][room_x]

    @staticmethod
    def make_pos(score, pos) -> MontezumaPosLevel:
        """Create a MontezumaPosLevel object using the provided data."""
        return MontezumaPosLevel(pos.level, score, pos.room, pos.x, pos.y)

    def render(self, mode="human", **kwargs) -> Union[None, numpy.ndarray]:
        """Render the environment."""
        return self.env.render(mode=mode)


# ------------------------------------------------------------------------------


class MontezumaEnv(AtariEnv):
    """Plangym implementation of the MontezumaEnv environment optimized for planning."""

    AVAILABLE_OBS_TYPES = {"coords", "rgb", "grayscale", "ram", None}

    def __init__(
        self,
        name="PlanMontezuma-v0",
        frameskip: int = 1,
        episodic_life: bool = False,
        autoreset: bool = True,
        delay_setup: bool = False,
        remove_time_limit: bool = True,
        obs_type: str = "rgb",  # coords | ram | rgb | grayscale
        mode: int = 0,  # game mode, see Machado et al. 2018
        difficulty: int = 0,  # game difficulty, see Machado et al. 2018
        repeat_action_probability: float = 0.0,  # Sticky action probability
        full_action_space: bool = False,  # Use all actions
        render_mode: Optional[str] = None,  # None | human | rgb_array
        possible_to_win: bool = True,
        wrappers: Iterable[wrap_callable] = None,
        array_state: bool = True,
        clone_seeds: bool = True,
        **kwargs,
    ):
        """Initialize a :class:`MontezumaEnv`."""
        super(MontezumaEnv, self).__init__(
            name="MontezumaRevengeDeterministic-v4",
            frameskip=frameskip,
            autoreset=autoreset,
            episodic_life=episodic_life,
            clone_seeds=clone_seeds,
            delay_setup=delay_setup,
            remove_time_limit=remove_time_limit,
            obs_type=obs_type,
            mode=mode,
            difficulty=difficulty,
            repeat_action_probability=repeat_action_probability,
            full_action_space=full_action_space,
            render_mode=render_mode,
            possible_to_win=possible_to_win,
            wrappers=wrappers,
            array_state=array_state,
            **kwargs,
        )

    def _get_default_obs_type(self, name, obs_type) -> str:
        value = super(MontezumaEnv, self)._get_default_obs_type(name, obs_type)
        if obs_type == "coords":
            value = obs_type
        return value

    def init_gym_env(self) -> CustomMontezuma:
        """Initialize the :class:`gum.Env`` instance that the current clas is wrapping."""
        kwargs = self._gym_env_kwargs
        kwargs["obs_type"] = self.obs_type
        return CustomMontezuma(**kwargs)

    def get_state(self) -> numpy.ndarray:
        """
        Recover the internal state of the simulation.

        If clone seed is False the environment will be stochastic.
        Cloning the full state ensures the environment is deterministic.
        """
        data = self.gym_env.get_restore()
        (
            full_state,
            score,
            steps,
            pos,
            room_time,
            ram_death_state,
            score_objects,
            cur_lives,
        ) = data
        room_time = room_time if room_time[0] is not None else (-1, -1)
        assert len(room_time) == 2
        metadata = numpy.array(
            [
                float(score),
                float(steps),
                float(room_time[0]),
                float(room_time[1]),
                float(ram_death_state),
                float(score_objects),
                float(cur_lives),
            ],
            dtype=float,
        )
        assert len(metadata) == 7
        posarray = numpy.array(pos.tuple, dtype=float)
        assert len(posarray) == 5
        array = numpy.concatenate([full_state, metadata, posarray]).astype(numpy.float32)
        return array

    def set_state(self, state: numpy.ndarray):
        """
        Set the internal state of the simulation.

        Args:
            state: Target state to be set in the environment.

        Returns:
            None
        """
        pos_vals = state[-5:].tolist()
        pos = MontezumaPosLevel(
            level=int(pos_vals[0]),
            score=float(pos_vals[1]),
            room=int(pos_vals[2]),
            x=float(pos_vals[3]),
            y=float(pos_vals[4]),
        )
        score, steps, rt0, rt1, ram_death_state, score_objects, cur_lives = state[-12:-5].tolist()
        room_time = (rt0, rt1) if rt0 != -1 and rt1 != -1 else (None, None)
        full_state = state[:-12].copy().astype(numpy.uint8)
        data = (
            full_state,
            score,
            steps,
            pos,
            room_time,
            int(ram_death_state),
            bool(score_objects),
            int(cur_lives),
        )
        self.gym_env.restore(data)



"""Implement the ``plangym`` API for ``gym`` classic control environments."""
import copy

import numpy

from plangym.core import PlangymEnv


class ClassicControl(PlangymEnv):
    """Environment for OpenAI gym classic control environments."""

    def get_state(self) -> numpy.ndarray:
        """Recover the internal state of the environment."""
        return numpy.array(copy.copy(self.gym_env.unwrapped.state))

    def set_state(self, state: numpy.ndarray):
        """
        Set the internal state of the environemnt.

        Args:
            state: Target state to be set in the environment.

        Returns:
            None

        """
        self.gym_env.unwrapped.state = copy.copy(tuple(state.tolist()))
        return state



import pytest

from plangym.vectorization.parallel import ParallelEnv
from plangym.videogames.montezuma import CustomMontezuma, MontezumaEnv, MontezumaPosLevel
from tests import SKIP_ATARI_TESTS


if SKIP_ATARI_TESTS:
    pytest.skip("Atari not installed, skipping", allow_module_level=True)
from plangym.api_tests import batch_size, display, TestPlanEnv, TestPlangymEnv  # noqa: F401


def montezuma():
    return MontezumaEnv(obs_type="coords", autoreset=True, score_objects=True)


def montezuma_unproc():
    return MontezumaEnv(obs_type="rgb", autoreset=True, only_keys=True)


def parallel_montezuma():
    return ParallelEnv(
        env_class=MontezumaEnv,
        frameskip=5,
        name="",
        score_objects=True,
        objects_from_pixels=True,
    )


def montezuma_coords():
    return MontezumaEnv(autoreset=True, obs_type="coords")


environments = [montezuma, montezuma_unproc, parallel_montezuma, montezuma_coords]


@pytest.fixture(params=environments, scope="module")
def env(request) -> MontezumaEnv:
    env = request.param()
    yield env
    env.close()


@pytest.fixture(scope="module")
def pos_level():
    return MontezumaPosLevel(1, 100, 2, 30, 16)


class TestMontezumaPosLevel:
    def test_hash(self, pos_level):
        assert isinstance(hash(pos_level), int)

    def test_compate(self, pos_level):
        assert pos_level.__eq__(MontezumaPosLevel(*pos_level.tuple))
        assert not pos_level == 6

    def test_get_state(self, pos_level):
        assert pos_level.__getstate__() == pos_level.tuple

    def test_set_state(self, pos_level):
        level, score, room, x, y = (10, 9, 8, 7, 6)
        pos_level.__setstate__((level, score, room, x, y))
        assert pos_level.tuple == (10, 9, 8, 7, 6)

    def test_repr(self, pos_level):
        assert isinstance(pos_level.__repr__(), str)


class TestCustomMontezuma:
    def test_make_pos(self, env):
        assert isinstance(env.gym_env.make_pos(1000, env.gym_env.pos), MontezumaPosLevel)

    def test_get_room(self):
        env = CustomMontezuma()
        env.get_room_xy(3)
        env.get_room_from_xy(0, 3)
        assert env.get_room_out_of_bounds(99, 99)
        assert not env.get_room_out_of_bounds(0, 0)

    def test_pos_from_unproc_state(self):
        env = CustomMontezuma(obs_type="rgb")
        obs = env.reset()
        for i in range(20):
            obs, *_ = env.step(0)
        facepix = env.get_face_pixels(obs)
        pos = env.pos_from_obs(face_pixels=facepix, obs=obs)
        assert isinstance(pos, MontezumaPosLevel)

    def test_get_objects_from_pixel(self):
        env = CustomMontezuma(obs_type="rgb")
        obs = env.reset()
        for i in range(20):
            obs, *_ = env.step(0)
        ob = env.get_objects_from_pixels(room=0, obs=obs, old_objects=[])
        assert isinstance(ob, int)

        env = CustomMontezuma(obs_type="rgb", objects_remember_rooms=True)
        obs = env.reset()
        for i in range(20):
            obs, *_ = env.step(0)
        tup = env.get_objects_from_pixels(room=0, obs=obs, old_objects=[])
        assert isinstance(tup, tuple)



import gym
from gym.wrappers.atari_preprocessing import AtariPreprocessing
from gym.wrappers.time_limit import TimeLimit
from gym.wrappers.transform_reward import TransformReward
import numpy

from plangym.utils import process_frame, remove_time_limit


def test_remove_time_limit():
    env = gym.make("MsPacmanNoFrameskip-v4")
    env = TransformReward(TimeLimit(AtariPreprocessing(env)), lambda x: x)
    rem_env = remove_time_limit(env)
    assert rem_env.spec.max_episode_steps == int(1e100)
    assert not isinstance(rem_env.env, TimeLimit)
    assert "TimeLimit" not in str(rem_env)


def test_process_frame():
    example = (numpy.random.random((100, 100, 3)) * 255).astype(numpy.uint8)
    frame = process_frame(example, mode="L")
    assert frame.shape == (100, 100)
    frame = process_frame(example, width=30, height=50)
    assert frame.shape == (50, 30, 3)
    frame = process_frame(example, width=80, height=70, mode="L")
    assert frame.shape == (70, 80)



import os

import pytest

from plangym.control.classic_control import ClassicControl
from plangym.environment_names import CLASSIC_CONTROL


if os.getenv("SKIP_RENDER", False) and str(os.getenv("SKIP_RENDER", False)).lower() != "false":
    pytest.skip("ClassicControl raises pyglet error on headless machines", allow_module_level=True)

from plangym.api_tests import (  # noqa: F401
    batch_size,
    display,
    generate_test_cases,
    TestPlanEnv,
    TestPlangymEnv,
)


@pytest.fixture(params=generate_test_cases(CLASSIC_CONTROL, ClassicControl), scope="module")
def env(request) -> ClassicControl:
    env = request.param()
    yield env
    env.close()



"""Implementation of LunarLander with no fire coming out of the engines that steps faster."""
import copy
import math
from typing import Iterable, Optional

import numpy

from plangym.control.box_2d import Box2DState
from plangym.core import PlangymEnv, wrap_callable


try:
    from Box2D.b2 import edgeShape, fixtureDef, polygonShape, revoluteJointDef
    from gym.envs.box2d.lunar_lander import ContactDetector, LunarLander as GymLunarLander

    import_error = None
except ImportError as e:
    import_error = e
    GymLunarLander = object

# Rocket trajectory optimization is a classic topic in Optimal Control.
#
# According to Pontryagin's maximum principle it's optimal to fire engine full throttle or
# turn it off. That's the reason this environment is OK to have discrete actions
# (engine on or off).
#
# Landing pad is always at coordinates (0,0). Coordinates are the first two
# numbers in the state vector.
# Reward for moving from the top of the screen to landing pad and zero speed is
# about 100..140 points.
# If lander moves away from landing pad it loses reward back. Episode finishes if the
# lander crashes or comes to rest, receiving additional -100 or +100 points.
# Each leg ground contact is +10.
# Firing main engine is -0.3 points each frame.
# Firing side engine is -0.03 points each frame. Solved is 200 points.
#
# Landing outside landing pad is possible. Fuel is infinite,
# so an agent can learn to fly and then land on its first attempt.
# Please see source code for details.
#
# To see heuristic landing, run:
#
# python gym/envs/box2d/lunar_lander.py
#
# To play yourself, run:
#
# python examples/agents/keyboard_agent.py LunarLander-v2
#
# Created by Oleg Klimov. Licensed on the same terms as the rest of OpenAI Gym.

FPS = 50
SCALE = 30.0  # affects how fast-paced the game is, forces should be adjusted as well

MAIN_ENGINE_POWER = 13.0
SIDE_ENGINE_POWER = 0.6

INITIAL_RANDOM = 1000.0  # Set 1500 to make game harder

LANDER_POLY = [(-14, +17), (-17, 0), (-17, -10), (+17, -10), (+17, 0), (+14, +17)]
LEG_AWAY = 20
LEG_DOWN = 18
LEG_W, LEG_H = 2, 8
LEG_SPRING_TORQUE = 40

SIDE_ENGINE_HEIGHT = 14.0
SIDE_ENGINE_AWAY = 12.0

VIEWPORT_W = 600
VIEWPORT_H = 400


class FastGymLunarLander(GymLunarLander):
    """Faster implementation of the LunarLander without bells and whistles."""

    FPS = FPS

    def __init__(self, deterministic: bool = False, continuous: bool = False):
        """Initialize a :class:`FastGymLunarLander``."""
        self.deterministic = deterministic
        self.game_over = False
        self.prev_shaping = None
        self.helipad_x1 = None
        self.helipad_x2 = None
        self.helipad_y = None
        self.moon = None
        self.sky_polys = None
        self.lander = None
        self.legs = None
        self.drawlist = None
        self.viewer = None
        self.moon = None
        self.lander = None
        self.particles = None
        self.prev_reward = None
        self.observation_space = None
        self.action_space = None
        self.continuous = continuous
        super(FastGymLunarLander, self).__init__()

    def reset(self) -> tuple:
        """Reset the environment to its initial state."""
        # Reset environment data
        self._destroy()
        self.world.contactListener_keepref = ContactDetector(self)
        self.world.contactListener = self.world.contactListener_keepref
        self.game_over = False
        self.prev_shaping = None
        # Define environment bodies
        W = VIEWPORT_W / SCALE
        H = VIEWPORT_H / SCALE
        # terrain shape
        CHUNKS = 11
        height = (
            numpy.ones(CHUNKS + 1) * H / 4
            if self.deterministic
            else self.np_random.uniform(0, H / 2, size=(CHUNKS + 1,))
        )
        # Define helipad
        chunk_x = [W / (CHUNKS - 1) * i for i in range(CHUNKS)]
        self.helipad_x1 = chunk_x[CHUNKS // 2 - 1]
        self.helipad_x2 = chunk_x[CHUNKS // 2 + 1]
        self.helipad_y = H / 4
        height[CHUNKS // 2 - 2] = self.helipad_y
        height[CHUNKS // 2 - 1] = self.helipad_y
        height[CHUNKS // 2 + 0] = self.helipad_y
        height[CHUNKS // 2 + 1] = self.helipad_y
        height[CHUNKS // 2 + 2] = self.helipad_y
        smooth_y = [0.33 * (height[i - 1] + height[i + 0] + height[i + 1]) for i in range(CHUNKS)]
        # Define moon
        self.moon = self.world.CreateStaticBody(shapes=edgeShape(vertices=[(0, 0), (W, 0)]))
        self.sky_polys = []
        for i in range(CHUNKS - 1):
            p1 = (chunk_x[i], smooth_y[i])
            p2 = (chunk_x[i + 1], smooth_y[i + 1])
            self.moon.CreateEdgeFixture(vertices=[p1, p2], density=0, friction=0.1)
            self.sky_polys.append([p1, p2, (p2[0], H), (p1[0], H)])
        self.moon.color1 = (0.0, 0.0, 0.0)
        self.moon.color2 = (0.0, 0.0, 0.0)
        # Define lander body and initial position
        initial_y = VIEWPORT_H / SCALE
        self.lander = self.world.CreateDynamicBody(
            position=(VIEWPORT_W / SCALE / 2, initial_y),
            angle=0.0,
            fixtures=fixtureDef(
                shape=polygonShape(vertices=[(x / SCALE, y / SCALE) for x, y in LANDER_POLY]),
                density=5.0,
                friction=0.1,
                categoryBits=0x0010,
                maskBits=0x001,  # collide only with ground
                restitution=0.0,
            ),  # 0.99 bouncy
        )
        self.lander.color1 = (0.5, 0.4, 0.9)
        self.lander.color2 = (0.3, 0.3, 0.5)
        # Unlike in the original LunarLander, the initial force can be deterministic.
        init_force_x = (
            0.0 if self.deterministic else self.np_random.uniform(-INITIAL_RANDOM, INITIAL_RANDOM)
        )
        init_force_y = (
            0.0 if self.deterministic else self.np_random.uniform(-INITIAL_RANDOM, INITIAL_RANDOM)
        )
        self.lander.ApplyForceToCenter((init_force_x, init_force_y), True)
        self.legs = []
        for i in [-1, +1]:
            leg = self.world.CreateDynamicBody(
                position=(VIEWPORT_W / SCALE / 2 - i * LEG_AWAY / SCALE, initial_y),
                angle=(i * 0.05),
                fixtures=fixtureDef(
                    shape=polygonShape(box=(LEG_W / SCALE, LEG_H / SCALE)),
                    density=1.0,
                    restitution=0.0,
                    categoryBits=0x0020,
                    maskBits=0x001,
                ),
            )
            leg.ground_contact = False
            leg.color1 = (0.5, 0.4, 0.9)
            leg.color2 = (0.3, 0.3, 0.5)
            rjd = revoluteJointDef(
                bodyA=self.lander,
                bodyB=leg,
                localAnchorA=(0, 0),
                localAnchorB=(i * LEG_AWAY / SCALE, LEG_DOWN / SCALE),
                enableMotor=True,
                enableLimit=True,
                maxMotorTorque=LEG_SPRING_TORQUE,
                motorSpeed=+0.3 * i,  # low enough not to jump back into the sky
            )
            if i == -1:
                # Yes, the most esoteric numbers here, angles legs have freedom to travel within
                rjd.lowerAngle = +0.9 - 0.5
                rjd.upperAngle = +0.9
            else:
                rjd.lowerAngle = -0.9
                rjd.upperAngle = -0.9 + 0.5
            leg.joint = self.world.CreateJoint(rjd)
            self.legs.append(leg)
        self.drawlist = [self.lander] + self.legs

        return self.step(numpy.array([0, 0]) if self.continuous else 0)[0]

    def step(self, action: int) -> tuple:
        """Step the environment applying the provided action."""
        if self.continuous:
            action = numpy.clip(action, -1, +1).astype(numpy.float32)
        else:
            assert self.action_space.contains(action), "%r (%s) invalid " % (action, type(action))

        # Engines
        tip = (math.sin(self.lander.angle), math.cos(self.lander.angle))
        side = (-tip[1], tip[0])
        dispersion = (
            [0, 0]
            if self.deterministic
            else [self.np_random.uniform(-1.0, +1.0) / SCALE for _ in range(2)]
        )
        # Main engine
        m_power = 0.0
        fire_me_continuous = self.continuous and action[0] > 0.0
        fire_me_discrete = not self.continuous and action == 2
        fire_main_engine = fire_me_continuous or fire_me_discrete
        if fire_main_engine:

            if self.continuous:
                m_power = (numpy.clip(action[0], 0.0, 1.0) + 1.0) * 0.5  # 0.5..1.0
                assert m_power >= 0.5 and m_power <= 1.0
            else:
                m_power = 1.0
            ox = (
                tip[0] * (4 / SCALE + 2 * dispersion[0]) + side[0] * dispersion[1]
            )  # 4 is move a bit downwards, +-2 for randomness
            oy = -tip[1] * (4 / SCALE + 2 * dispersion[0]) - side[1] * dispersion[1]
            impulse_pos = (self.lander.position[0] + ox, self.lander.position[1] + oy)
            # We do not create any decorative particle.
            self.lander.ApplyLinearImpulse(
                (-ox * MAIN_ENGINE_POWER * m_power, -oy * MAIN_ENGINE_POWER * m_power),
                impulse_pos,
                True,
            )
        # Orientation engines
        s_power = 0.0
        fire_oe_continuous = self.continuous and numpy.abs(action[1]) > 0.5
        fire_oe_discrete = not self.continuous and action in [1, 3]
        fire_orientation_engine = fire_oe_continuous or fire_oe_discrete
        if fire_orientation_engine:
            if self.continuous:
                direction = numpy.sign(action[1])
                s_power = numpy.clip(numpy.abs(action[1]), 0.5, 1.0)
                assert s_power >= 0.5 and s_power <= 1.0
            else:
                direction = action - 2
                s_power = 1.0
            ox = tip[0] * dispersion[0] + side[0] * (
                3 * dispersion[1] + direction * SIDE_ENGINE_AWAY / SCALE
            )
            oy = -tip[1] * dispersion[0] - side[1] * (
                3 * dispersion[1] + direction * SIDE_ENGINE_AWAY / SCALE
            )
            impulse_pos = (
                self.lander.position[0] + ox - tip[0] * 17 / SCALE,
                self.lander.position[1] + oy + tip[1] * SIDE_ENGINE_HEIGHT / SCALE,
            )
            self.lander.ApplyLinearImpulse(
                (-ox * SIDE_ENGINE_POWER * s_power, -oy * SIDE_ENGINE_POWER * s_power),
                impulse_pos,
                True,
            )

        self.world.Step(1.0 / FPS, 6 * 30, 2 * 30)

        pos = self.lander.position
        vel = self.lander.linearVelocity
        state = [
            (pos.x - VIEWPORT_W / SCALE / 2) / (VIEWPORT_W / SCALE / 2),
            (pos.y - (self.helipad_y + LEG_DOWN / SCALE)) / (VIEWPORT_H / SCALE / 2),
            vel.x * (VIEWPORT_W / SCALE / 2) / FPS,
            vel.y * (VIEWPORT_H / SCALE / 2) / FPS,
            self.lander.angle,
            20.0 * self.lander.angularVelocity / FPS,
            1.0 if self.legs[0].ground_contact else 0.0,
            1.0 if self.legs[1].ground_contact else 0.0,
        ]
        assert len(state) == 8

        reward = 0
        shaping = (
            -100 * numpy.sqrt(state[0] * state[0] + state[1] * state[1])
            - 100 * numpy.sqrt(state[2] * state[2] + state[3] * state[3])
            - 100 * abs(state[4])
            + 10 * state[6]
            + 10 * state[7]
        )  # And ten points for legs contact, the idea is if you
        # lose contact again after landing, you get negative reward
        if self.prev_shaping is not None:
            reward = shaping - self.prev_shaping
        self.prev_shaping = shaping

        reward -= m_power * 0.30  # less fuel spent is better, about -30 for heuristic landing
        reward -= s_power * 0.03

        done = False
        if self.game_over or abs(state[0]) >= 1.0:
            done = True
            reward = -100
        if not self.lander.awake:  # pragma: no cover
            done = True
            reward = +100
        self.prev_reward = reward
        self.game_over = done or self.game_over
        return numpy.array(state, dtype=numpy.float32), reward, done, {}

    def render(self, mode="human"):
        """Render the environment."""
        from gym.envs.classic_control import rendering

        if self.viewer is None:
            self.viewer = rendering.Viewer(VIEWPORT_W, VIEWPORT_H)
            self.viewer.set_bounds(0, VIEWPORT_W / SCALE, 0, VIEWPORT_H / SCALE)

        for p in self.sky_polys:
            self.viewer.draw_polygon(p, color=(0, 0, 0))

        for obj in self.drawlist:
            for f in obj.fixtures:
                trans = f.body.transform
                path = [trans * v for v in f.shape.vertices]
                self.viewer.draw_polygon(path, color=obj.color1)
                path.append(path[0])
                self.viewer.draw_polyline(path, color=obj.color2, linewidth=2)

        for x in [self.helipad_x1, self.helipad_x2]:
            flagy1 = self.helipad_y
            flagy2 = flagy1 + 50 / SCALE
            self.viewer.draw_polyline([(x, flagy1), (x, flagy2)], color=(1, 1, 1))
            self.viewer.draw_polygon(
                [
                    (x, flagy2),
                    (x, flagy2 - 10 / SCALE),
                    (x + 25 / SCALE, flagy2 - 5 / SCALE),
                ],
                color=(0.8, 0.8, 0),
            )

        return self.viewer.render(return_rgb_array=mode == "rgb_array")


class LunarLander(PlangymEnv):
    """Fast LunarLander that follows the plangym API."""

    def __init__(
        self,
        name: str = None,
        frameskip: int = 1,
        episodic_life: bool = True,
        autoreset: bool = True,
        wrappers: Iterable[wrap_callable] = None,
        delay_setup: bool = False,
        deterministic: bool = False,
        continuous: bool = False,
        render_mode: Optional[str] = None,
        remove_time_limit=None,
        **kwargs,
    ):
        """Initialize a :class:`LunarLander`."""
        self._deterministic = deterministic
        self._continuous = continuous
        super(LunarLander, self).__init__(
            name="LunarLander-plangym",
            frameskip=frameskip,
            episodic_life=episodic_life,
            autoreset=autoreset,
            wrappers=wrappers,
            delay_setup=delay_setup,
            render_mode=render_mode,
            **kwargs,
        )

    @property
    def deterministic(self) -> bool:
        """Return true if the LunarLander simulation is deterministic."""
        return self._deterministic

    @property
    def continuous(self) -> bool:
        """Return true if the LunarLander agent takes continuous actions as input."""
        return self._continuous

    def init_gym_env(self) -> FastGymLunarLander:
        """Initialize the target :class:`gym.Env` instance."""
        if import_error is not None:
            raise import_error
        gym_env = FastGymLunarLander(
            deterministic=self.deterministic,
            continuous=self.continuous,
        )
        gym_env.reset()
        return gym_env

    def get_state(self) -> numpy.ndarray:
        """
        Recover the internal state of the simulation.

        An state must completely describe the Environment at a given moment.
        """
        env_data = [
            bool(self.gym_env.lander.awake),
            bool(self.gym_env.game_over),
            copy.copy(self.gym_env.prev_shaping),
            copy.copy(self.gym_env.prev_reward),
            bool(self.gym_env.legs[0].ground_contact),
            bool(self.gym_env.legs[1].ground_contact),
        ]
        state = Box2DState.get_env_state(self.gym_env) + env_data
        return numpy.array((state, None), dtype=object)

    def set_state(self, state: numpy.ndarray) -> None:
        """
        Set the internal state of the simulation.

        Args:
            state: Target state to be set in the environment.

        Returns:
            None

        """
        box_2d_state = state[0][:-6]
        Box2DState.set_env_state(self.gym_env, box_2d_state)
        self.gym_env.lander.awake = state[0][-6]
        self.gym_env.game_over = state[0][-5]
        self.gym_env.prev_shaping = state[0][-4]
        self.gym_env.prev_reward = state[0][-3]
        self.gym_env.legs[0].ground_contact = state[0][-2]
        self.gym_env.legs[1].ground_contact = state[0][-1]

    def process_terminal(self, terminal, obs=None, **kwargs) -> bool:
        """Return the terminal condition considering the lunar lander state."""
        obs = [0] if obs is None else obs
        end = (
            self.gym_env.game_over
            or (self.obs_type == "coords" and abs(obs[0]) >= 1.0)
            or not self.gym_env.lander.awake
        )
        return terminal or end



"""Implement a :class:`plangym.VectorizedEnv` that uses ray when calling `step_batch`."""
from typing import List

import numpy


try:
    import ray
except ImportError:
    pass

from plangym.core import PlanEnv
from plangym.vectorization.env import VectorizedEnv


@ray.remote
class RemoteEnv(PlanEnv):
    """Remote ray Actor interface for a plangym.PlanEnv."""

    def __init__(self, env_callable):
        """Initialize a :class:`RemoteEnv`."""
        self._env_callable = env_callable
        self.env = None

    @property
    def unwrapped(self):
        """Completely unwrap this Environment.

        Returns:
            plangym.Environment: The base non-wrapped plangym.Environment instance
        """
        return self.env

    @property
    def name(self) -> str:
        """Return the name of the environment."""
        return self.env.name

    def setup(self):
        """Init the wrapped environment."""
        self.env = self._env_callable()

    def step(self, action, state=None, dt: int = 1, return_state: bool = None) -> tuple:
        """
        Take a simulation step and make the environment evolve.

        Args:
            action: Chosen action applied to the environment.
            state: Set the environment to the given state before stepping it.
                If state is None the behaviour of this function will be the
                same as in OpenAI gym.
            dt: Consecutive number of times to apply an action.
            return_state: Whether to return the state in the returned tuple. \
                If None, `step` will return the state if `state` was passed as a parameter.

        Returns:
            if states is None returns (observs, rewards, ends, infos)
            else returns(new_states, observs, rewards, ends, infos)
        """
        return self.env.step(action=action, state=state, dt=dt, return_state=return_state)

    def step_batch(
        self,
        actions: [numpy.ndarray, list],
        states=None,
        dt: int = 1,
        return_state: bool = None,
    ) -> tuple:
        """
        Take a step on a batch of states and actions.

        Args:
            actions: Chosen actions applied to the environment.
            states: Set the environment to the given states before stepping it.
                If state is None the behaviour of this function will be the same
                as in OpenAI gym.
            dt: Consecutive number of times that the action will be
                applied.
            return_state: Whether to return the state in the returned tuple. \
                If None, `step` will return the state if `state` was passed as a parameter.

        Returns:
            if states is None returns (observs, rewards, ends, infos)
            else returns(new_states, observs, rewards, ends, infos)
        """
        return self.env.step_batch(
            actions=actions,
            states=states,
            dt=dt,
            return_state=return_state,
        )

    def reset(self, return_state: bool = True) -> [numpy.ndarray, tuple]:
        """Restart the environment."""
        return self.env.reset(return_state=return_state)

    def get_state(self):
        """
        Recover the internal state of the simulation.

        A state must completely describe the Environment at a given moment.
        """
        return self.env.get_state()

    def set_state(self, state):
        """
        Set the internal state of the simulation.

        Args:
            state: Target state to be set in the environment.

        Returns:
            None
        """
        return self.env.set_state(state=state)


class RayEnv(VectorizedEnv):
    """Use ray for taking steps in parallel when calling `step_batch`."""

    def __init__(
        self,
        env_class,
        name: str,
        frameskip: int = 1,
        autoreset: bool = True,
        delay_setup: bool = False,
        n_workers: int = 8,
        **kwargs,
    ):
        """
        Initialize a :class:`ParallelEnv`.

        Args:
            env_class: Class of the environment to be wrapped.
            name: Name of the environment.
            frameskip: Number of times ``step`` will me called with the same action.
            autoreset: Ignored. Always set to True. Automatically reset the environment
                      when the OpenAI environment returns ``end = True``.
            delay_setup: If ``True`` do not initialize the ``gym.Environment`` \
                     and wait for ``setup`` to be called later.
            env_callable: Callable that returns an instance of the environment \
                         that will be parallelized.
            n_workers:  Number of workers that will be used to step the env.
            *args: Additional args for the environment.
            **kwargs: Additional kwargs for the environment.

        """
        self._workers = None
        super(RayEnv, self).__init__(
            env_class=env_class,
            name=name,
            frameskip=frameskip,
            autoreset=autoreset,
            delay_setup=delay_setup,
            n_workers=n_workers,
            **kwargs,
        )

    @property
    def workers(self) -> List[RemoteEnv]:
        """Remote actors exposing copies of the environment."""
        return self._workers

    def setup(self):
        """Run environment initialization and create the subprocesses for stepping in parallel."""
        env_callable = self.create_env_callable(autoreset=True, delay_setup=False)
        workers = [RemoteEnv.remote(env_callable=env_callable) for _ in range(self.n_workers)]
        ray.get([w.setup.remote() for w in workers])
        self._workers = workers
        # Initialize local copy last to tolerate singletons better
        super(RayEnv, self).setup()

    def make_transitions(
        self,
        actions,
        states=None,
        dt: [numpy.ndarray, int] = 1,
        return_state: bool = None,
    ):
        """Implement the logic for stepping the environment in parallel."""
        no_states = states is None or states[0] is None
        _return_state = ((not no_states) and return_state is None) or return_state
        chunks = self.batch_step_data(
            actions=actions,
            states=states,
            dt=dt,
            batch_size=len(self.workers),
        )
        results_ids = []
        for env, states_batch, actions_batch, dt in zip(self.workers, *chunks):
            result = env.step_batch.remote(
                actions=actions_batch,
                states=states_batch,
                dt=dt,
                return_state=return_state,
            )
            results_ids.append(result)
        results = ray.get(results_ids)
        return self.unpack_transitions(results=results, return_states=_return_state)

    def reset(self, return_state: bool = True) -> [numpy.ndarray, tuple]:
        """Restart the environment."""
        if self.plan_env is None and self.delay_setup:
            self.setup()
        ray.get([w.reset.remote(return_state=return_state) for w in self.workers])
        return super(RayEnv, self).reset(return_state=return_state)

    def sync_states(self, state: None) -> None:
        """
        Synchronize all the copies of the wrapped environment.

        Set all the states of the different workers of the internal :class:`BatchEnv`
         to the same state as the internal :class:`Environment` used to apply the
         non-vectorized steps.
        """
        state = super().get_state() if state is None else state
        obj_ids = [w.set_state.remote(state) for w in self.workers]
        self.plan_env.set_state(state)
        ray.get(obj_ids)



"""Various environments for plangym."""
import warnings


warnings.filterwarnings(
    "ignore",
    message=(
        "Using or importing the ABCs from 'collections' instead of from 'collections.abc' "
        "is deprecated since Python 3.3,and in 3.9 it will stop working"
    ),
)
warnings.filterwarnings(
    "ignore",
    message=(
        "the imp module is deprecated in favour of importlib; see the module's "
        "documentation for alternative uses"
    ),
)
warnings.filterwarnings(
    "ignore",
    message=(
        "Using or importing the ABCs from 'collections' instead of from "
        "'collections.abc' is deprecated, and in 3.8 it will stop working"
    ),
)
warnings.filterwarnings(
    "ignore",
    message=(
        "The set_clim function was deprecated in Matplotlib 3.1 "
        "and will be removed in 3.3. Use ScalarMappable.set_clim "
        "instead."
    ),
)
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
)
warnings.filterwarnings(
    "ignore",
    message="invalid escape sequence",
)

warnings.filterwarnings("ignore", message="Gdk.Cursor.new is deprecated")
warnings.filterwarnings(
    "ignore",
    message=(
        " `numpy.bool` is a deprecated alias for the builtin `bool`. "
        "To silence this warning, use `bool` by itself. Doing this will not modify any "
        "behavior and is safe. If you specifically wanted the numpy scalar type, "
        "use `numpy.bool_` here."
    ),
)
warnings.filterwarnings(
    "ignore",
    message=" WARN: Box bound precision lowered by casting to float32",
)
warnings.filterwarnings(
    "ignore",
    message=(
        " DeprecationWarning: The binary mode of fromstring is deprecated, "
        "as it behaves surprisingly on unicode inputs. Use frombuffer instead"
    ),
)
warnings.filterwarnings(
    "ignore",
    message=(
        " DeprecationWarning: distutils Version classes are deprecated. "
        "Use packaging.version instead."
    ),
)
warnings.filterwarnings(
    "ignore",
    message=" WARNING:root:The use of `check_types` is deprecated and does not have any effect.",
)

from plangym.core import PlanEnv  # noqa: E402
from plangym.registry import make  # noqa: E402
from plangym.version import __version__  # noqa: E402



"""Implement the ``plangym`` API for the Balloon Learning Environment."""
from typing import Any

import numpy


try:
    import balloon_learning_environment.env.balloon_env  # noqa: F401
    from balloon_learning_environment.env.rendering.matplotlib_renderer import MatplotlibRenderer
except ImportError:

    def MatplotlibRenderer():  # noqa: D103
        return None


from plangym.core import PlangymEnv


class BalloonEnv(PlangymEnv):
    """
    This class implements the 'BalloonLearningEnvironment-v0' released by Google in the \
    balloon_learning_environment.

    For more information about the environment, please refer to \
    https://github.com/google/balloon-learning-environment.
    """

    AVAILABLE_RENDER_MODES = {"human", "rgb_array", "tensorboard", None}
    AVAILABLE_OBS_TYPES = {"coords", "rgb", "grayscale"}
    STATE_IS_ARRAY = False

    def __init__(
        self,
        name: str = "BalloonLearningEnvironment-v0",
        renderer=None,
        array_state: bool = True,
        **kwargs,
    ):
        """
        Initialize a :class:`BalloonEnv`.

        Args:
            name: Name of the environment. Follows standard gym syntax conventions.
            renderer: MatplotlibRenderer object (or any renderer object) to plot
                the ``balloons`` environment. For more information, see the
                official documentation.
            array_state: boolean value. If True, transform the state object to
                a ``numpy.array``.
        """
        renderer = renderer or MatplotlibRenderer()
        self.STATE_IS_ARRAY = array_state
        super(BalloonEnv, self).__init__(name=name, renderer=renderer, **kwargs)

    def get_state(self) -> Any:
        """Get the state of the environment."""
        state = self.gym_env.unwrapped.get_simulator_state()
        if self.STATE_IS_ARRAY:
            state = numpy.array((state, None), dtype=object)
        return state

    def set_state(self, state: Any) -> None:
        """Set the state of the environment."""
        if self.STATE_IS_ARRAY:
            state = state[0]
        return self.gym_env.unwrapped.arena.set_simulator_state(state)

    def seed(self, seed: int = None):
        """Ignore seeding until next release."""
        return



"""Implement the ``plangym`` API for Atari environments."""
from typing import Any, Dict, Iterable, Optional, Union

import gym
from gym.spaces import Space
import numpy

from plangym.core import wrap_callable
from plangym.videogames.env import VideogameEnv


def ale_to_ram(ale) -> numpy.ndarray:
    """Return the ram of the ale emulator."""
    ram_size = ale.getRAMSize()
    ram = numpy.zeros(ram_size, dtype=numpy.uint8)
    ale.getRAM(ram)
    return ram


class AtariEnv(VideogameEnv):
    """
    Create an environment to play OpenAI gym Atari Games that uses AtariALE as the emulator.

    Args:
        name: Name of the environment. Follows standard gym syntax conventions.
        frameskip: Number of times an action will be applied for each step
            in dt.
        episodic_life: Return ``end = True`` when losing a life.
        autoreset: Restart environment when reaching a terminal state.
        delay_setup: If ``True`` do not initialize the ``gym.Environment``
            and wait for ``setup`` to be called later.
        remove_time_limit: If True, remove the time limit from the environment.
        obs_type: One of {"rgb", "ram", "grayscale"}.
        mode: Integer or string indicating the game mode, when available.
        difficulty: Difficulty level of the game, when available.
        repeat_action_probability: Repeat the last action with this probability.
        full_action_space: Wheter to use the full range of possible actions
            or only those available in the game.
        render_mode: One of {None, "human", "rgb_aray"}.
        possible_to_win: It is possible to finish the Atari game without getting a
            terminal state that is not out of bounds or does not involve losing a life.
        wrappers: Wrappers that will be applied to the underlying OpenAI env.
            Every element of the iterable can be either a :class:`gym.Wrapper`
            or a tuple containing ``(gym.Wrapper, kwargs)``.
        array_state: Whether to return the state of the environment as a numpy array.
        clone_seeds: Clone the random seed of the ALE emulator when reading/setting
            the state. False makes the environment stochastic.

    Example::

        >>> env = plangym.make(name="ALE/MsPacman-v5", difficulty=2, mode=1)
        >>> state, obs = env.reset()
        >>>
        >>> states = [state.copy() for _ in range(10)]
        >>> actions = [env.action_space.sample() for _ in range(10)]
        >>>
        >>> data = env.step_batch(states=states, actions=actions)
        >>> new_states, observs, rewards, ends, infos = data

    """

    STATE_IS_ARRAY = True

    def __init__(
        self,
        name: str,
        frameskip: int = 5,
        episodic_life: bool = False,
        autoreset: bool = True,
        delay_setup: bool = False,
        remove_time_limit: bool = True,
        obs_type: str = "rgb",  # ram | rgb | grayscale
        mode: int = 0,  # game mode, see Machado et al. 2018
        difficulty: int = 0,  # game difficulty, see Machado et al. 2018
        repeat_action_probability: float = 0.0,  # Sticky action probability
        full_action_space: bool = False,  # Use all actions
        render_mode: Optional[str] = None,  # None | human | rgb_array
        possible_to_win: bool = False,
        wrappers: Iterable[wrap_callable] = None,
        array_state: bool = True,
        clone_seeds: bool = False,
        **kwargs,
    ):
        """
        Initialize a :class:`AtariEnvironment`.

        Args:
            name: Name of the environment. Follows standard gym syntax conventions.
            frameskip: Number of times an action will be applied for each step
                in dt.
            episodic_life: Return ``end = True`` when losing a life.
            autoreset: Restart environment when reaching a terminal state.
            delay_setup: If ``True`` do not initialize the ``gym.Environment``
                and wait for ``setup`` to be called later.
            remove_time_limit: If True, remove the time limit from the environment.
            obs_type: One of {"rgb", "ram", "grayscale"}.
            mode: Integer or string indicating the game mode, when available.
            difficulty: Difficulty level of the game, when available.
            repeat_action_probability: Repeat the last action with this probability.
            full_action_space: Wheter to use the full range of possible actions
                or only those available in the game.
            render_mode: One of {None, "human", "rgb_aray"}.
            possible_to_win: It is possible to finish the Atari game without getting a
                terminal state that is not out of bounds or does not involve losing a life.
            wrappers: Wrappers that will be applied to the underlying OpenAI env.
                Every element of the iterable can be either a :class:`gym.Wrapper`
                or a tuple containing ``(gym.Wrapper, kwargs)``.
            array_state: Whether to return the state of the environment as a numpy array.
            clone_seeds: Clone the random seed of the ALE emulator when reading/setting
                the state. False makes the environment stochastic.

        Example::

            >>> env = AtariEnv(name="ALE/MsPacman-v5", difficulty=2, mode=1)
            >>> type(env.gym_env)
            <class 'gym.envs.atari.environment.AtariEnv'>
            >>> state, obs = env.reset()
            >>> type(state)
            <class 'numpy.ndarray'>

        """
        self.clone_seeds = clone_seeds
        self._mode = mode
        self._difficulty = difficulty
        self._repeat_action_probability = repeat_action_probability
        self._full_action_space = full_action_space
        self.STATE_IS_ARRAY = array_state
        self.DEFAULT_OBS_TYPE = self._get_default_obs_type(name, obs_type)
        super(AtariEnv, self).__init__(
            name=name,
            frameskip=frameskip,
            episodic_life=episodic_life,
            autoreset=autoreset,
            delay_setup=delay_setup,
            remove_time_limit=remove_time_limit,
            obs_type=obs_type,  # ram | rgb | grayscale
            render_mode=render_mode,  # None | human | rgb_array
            wrappers=wrappers,
            **kwargs,
        )

    @property
    def ale(self):
        """
        Return the ``ale`` interface of the underlying :class:`gym.Env`.

        Example::

            >>> env = AtariEnv(name="ALE/MsPacman-v5", obs_type="ram")
            >>> type(env.ale)
            <class 'ale_py._ale_py.ALEInterface'>


        """
        return self.gym_env.unwrapped.ale

    @property
    def mode(self) -> int:
        """Return the selected game mode for the current environment."""
        return self._mode

    @property
    def difficulty(self) -> int:
        """Return the selected difficulty for the current environment."""
        return self._difficulty

    @property
    def repeat_action_probability(self) -> float:
        """Probability of repeating the same action after input."""
        return self._repeat_action_probability

    @property
    def full_action_space(self) -> bool:
        """If True the action space correspond to all possible actions in the Atari emulator."""
        return self._full_action_space

    @property
    def observation_space(self) -> Space:
        """Return the observation_space of the environment."""
        return self.gym_env.observation_space

    @staticmethod
    def _get_default_obs_type(name, obs_type) -> str:
        """Return the observation type of the internal Atari gym environment."""
        if "ram" in name or obs_type == "ram":
            return "ram"
        elif obs_type == "grayscale":
            return "grayscale"
        return "rgb"

    def get_lifes_from_info(self, info: Dict[str, Any]) -> int:
        """Return the number of lives remaining in the current game."""
        return info.get("ale.lives", super().get_lifes_from_info(info))

    def get_image(self) -> numpy.ndarray:
        """
        Return a numpy array containing the rendered view of the environment.

        Image is a three-dimensional array interpreted as an RGB image with
        channels (Height, Width, RGB). Ignores wrappers as it loads the
        screen directly from the emulator.

        Example::

            >>> env = AtariEnv(name="ALE/MsPacman-v5", obs_type="ram")
            >>> img = env.get_image()
            >>> img.shape
            (210, 160, 3)
        """
        return self.gym_env.ale.getScreenRGB()

    def get_ram(self) -> numpy.ndarray:
        """
        Return a numpy array containing the content of the emulator's RAM.

        The RAM is a vector array interpreted as the memory of the emulator.

         Example::

            >>> env = AtariEnv(name="ALE/MsPacman-v5", obs_type="grayscale")
            >>> ram = env.get_ram()
            >>> ram.shape, ram.dtype
            ((128,), dtype('uint8'))
        """
        return self.gym_env.ale.getRAM()

    def init_gym_env(self) -> gym.Env:
        """Initialize the :class:`gym.Env`` instance that the Environment is wrapping."""
        # Remove any undocumented wrappers
        try:
            default_env_kwargs = dict(
                obs_type=self.obs_type,  # ram | rgb | grayscale
                frameskip=self.frameskip,  # frame skip
                mode=self._mode,  # game mode, see Machado et al. 2018
                difficulty=self.difficulty,  # game difficulty, see Machado et al. 2018
                repeat_action_probability=self.repeat_action_probability,  # Sticky action prob
                full_action_space=self.full_action_space,  # Use all actions
                render_mode=self.render_mode,  # None | human | rgb_array
            )
            default_env_kwargs.update(self._gym_env_kwargs)
            self._gym_env_kwargs = default_env_kwargs
            gym_env = super(AtariEnv, self).init_gym_env()
        except RuntimeError:
            gym_env: gym.Env = gym.make(self.name)
            gym_env.reset()
        return gym_env

    def get_state(self) -> numpy.ndarray:
        """
        Recover the internal state of the simulation.

        If clone seed is False the environment will be stochastic.
        Cloning the full state ensures the environment is deterministic.

        Example::

            >>> env = AtariEnv(name="Qbert-v0")
            >>> env.get_state() #doctest: +ELLIPSIS
            array([<ale_py._ale_py.ALEState object at 0x...>, None],
                  dtype=object)

            >>> env = AtariEnv(name="Qbert-v0", array_state=False)
            >>> env.get_state() #doctest: +ELLIPSIS
            <ale_py._ale_py.ALEState object at 0x...>

        """
        state = self.gym_env.unwrapped.clone_state(include_rng=self.clone_seeds)
        if self.STATE_IS_ARRAY:
            state = numpy.array((state, None), dtype=object)
        return state

    def set_state(self, state: numpy.ndarray) -> None:
        """
        Set the internal state of the simulation.

        Args:
            state: Target state to be set in the environment.

        Example::

            >>> env = AtariEnv(name="Qbert-v0")
            >>> state, obs = env.reset()
            >>> new_state, obs, reward, end, info = env.step(env.sample_action(), state=state)
            >>> assert not (state == new_state).all()
            >>> env.set_state(state)
            >>> (state == env.get_state()).all()
            True
        """
        if self.STATE_IS_ARRAY:
            state = state[0]
        self.gym_env.unwrapped.restore_state(state)

    def step_with_dt(self, action: Union[numpy.ndarray, int, float], dt: int = 1):
        """
        Take ``dt`` simulation steps and make the environment evolve in multiples \
        of ``self.frameskip`` for a total of ``dt`` * ``self.frameskip`` steps.

        Args:
            action: Chosen action applied to the environment.
            dt: Consecutive number of times that the action will be applied.

        Returns:
            If state is `None` return ``(observs, reward, terminal, info)``
            else returns ``(new_state, observs, reward, terminal, info)``

        Example::

            >>> env = AtariEnv(name="Pong-v0")
            >>> obs = env.reset(return_state=False)
            >>> obs, reward, end, info = env.step_with_dt(env.sample_action(), dt=7)
            >>> assert not end

        """
        return super(AtariEnv, self).step_with_dt(action=action, dt=dt)

    def clone(self, **kwargs) -> "VideogameEnv":
        """Return a copy of the environment."""
        params = dict(
            mode=self.mode,
            difficulty=self.difficulty,
            repeat_action_probability=self.repeat_action_probability,
            full_action_space=self.full_action_space,
        )
        params.update(**kwargs)
        return super(VideogameEnv, self).clone(**params)


class AtariPyEnvironment(AtariEnv):
    """Create an environment to play OpenAI gym Atari Games that uses AtariPy as the emulator."""

    def get_state(self) -> numpy.ndarray:  # pragma: no cover
        """
        Recover the internal state of the simulation.

        If clone seed is False the environment will be stochastic.
        Cloning the full state ensures the environment is deterministic.
        """
        if self.clone_seeds:
            return self.gym_env.unwrapped.clone_full_state()
        else:
            return self.gym_env.unwrapped.clone_state()

    def set_state(self, state: numpy.ndarray) -> None:  # pragma: no cover
        """
        Set the internal state of the simulation.

        Args:
            state: Target state to be set in the environment.

        Returns:
            None

        """
        state = state.astype(numpy.uint8)
        if self.clone_seeds:
            self.gym_env.unwrapped.restore_full_state(state)
        else:
            self.gym_env.unwrapped.restore_state(state)

    def get_ram(self) -> numpy.ndarray:  # pragma: no cover
        """
        Return a numpy array containing the content of the emulator's RAM.

        The RAM is a vector array interpreted as the memory of the emulator.
        """
        return ale_to_ram(self.gym_env.unwrapped.ale)



import pytest


pytest.importorskip("Box2D")
from plangym.api_tests import (  # noqa: F401
    batch_size,
    display,
    generate_test_cases,
    TestPlanEnv,
    TestPlangymEnv,
)
from plangym.control.lunar_lander import FastGymLunarLander, LunarLander


def lunar_lander_det_discrete():
    return LunarLander(autoreset=False, deterministic=True, continuous=False)


def lunar_lander_random_discrete():
    return LunarLander(autoreset=False, deterministic=False, continuous=False)


def lunar_lander_random_continuous():
    return LunarLander(
        autoreset=False,
        deterministic=False,
        continuous=True,
    )


environments = [
    lunar_lander_det_discrete,
    lunar_lander_random_discrete,
    lunar_lander_random_continuous,
]


@pytest.fixture(
    params=generate_test_cases(["FastLunarLander-v0"], LunarLander, custom_tests=environments),
    scope="module",
)
def env(request) -> LunarLander:
    env = request.param()
    yield env
    env.close()


class TestFastGymLunarLander:
    def test_death(self):
        gym_env = FastGymLunarLander()
        gym_env.reset()
        for _ in range(1000):
            *_, end, info = gym_env.step(gym_env.action_space.sample())
            if end:
                break



import pytest

from plangym.api_tests import (  # noqa: F401
    batch_size,
    display,
    generate_test_cases,
    TestPlanEnv,
    TestPlangymEnv,
    TestVideogameEnv,
)
from plangym.videogames.nes import MarioEnv


env_names = ["SuperMarioBros-v0", "SuperMarioBros-v1", "SuperMarioBros2-v0"]


@pytest.fixture(
    params=generate_test_cases(env_names, MarioEnv, n_workers_values=[None, 2]), scope="module"
)
def env(request):
    return request.param()



"""Handle parallelization for ``plangym.Environment`` that allows vectorized steps."""
import atexit
import multiprocessing
import sys
import traceback
from typing import Union

import numpy

from plangym.core import PlanEnv
from plangym.vectorization.env import VectorizedEnv


class ExternalProcess:
    """
    Step environment in a separate process for lock free paralellism.

    The environment will be created in the external process by calling the
    specified callable. This can be an environment class, or a function
    creating the environment and potentially wrapping it. The returned
    environment should not access global variables.

    Args:
      constructor: Callable that creates and returns an OpenAI gym environment.

    Attributes:
      observation_space: The cached observation space of the environment.
      action_space: The cached action space of the environment.

    ..notes:
        This is mostly a copy paste from
        https://github.com/tensorflow/agents/blob/master/agents/tools/wrappers.py,
        that lets us set and read the environment state.

    """

    # Message types for communication via the pipe.
    _ACCESS = 1
    _CALL = 2
    _RESULT = 3
    _EXCEPTION = 4
    _CLOSE = 5

    def __init__(self, constructor):
        """
        Initialize a :class:`ExternalProcess`.

        Args:
            constructor: Callable that returns the target environment that will be parallelized.

        """
        self._conn, conn = multiprocessing.Pipe()
        self._process = multiprocessing.Process(target=self._worker, args=(constructor, conn))
        atexit.register(self.close)
        self._process.start()
        self._observ_space = None
        self._action_space = None

    @property
    def observation_space(self):
        """Return the observation space of the internal environment."""
        if not self._observ_space:
            self._observ_space = self.__getattr__("observation_space")
        return self._observ_space

    @property
    def action_space(self):
        """Return the action space of the internal environment."""
        if not self._action_space:
            self._action_space = self.__getattr__("action_space")
        return self._action_space

    def __getattr__(self, name):
        """
        Request an attribute from the environment.

        Note that this involves communication with the external process, so it can \
        be slow.

        Args:
          name: Attribute to access.

        Returns:
          Value of the attribute.

        """
        self._conn.send((self._ACCESS, name))
        return self._receive()

    def call(self, name, *args, **kwargs):
        """
        Asynchronously call a method of the external environment.

        Args:
          name: Name of the method to call.
          *args: Positional arguments to forward to the method.
          **kwargs: Keyword arguments to forward to the method.

        Returns:
          Promise object that blocks and provides the return value when called.

        """
        payload = name, args, kwargs
        self._conn.send((self._CALL, payload))
        return self._receive

    def close(self):
        """Send a close message to the external process and join it."""
        try:
            self._conn.send((self._CLOSE, None))
            self._conn.close()
        except IOError:
            # The connection was already closed.
            pass
        self._process.join()

    def set_state(self, state, blocking=True):
        """Set the state of the internal environment."""
        promise = self.call("set_state", state)
        return promise() if blocking else promise

    def step_batch(
        self,
        actions,
        states=None,
        dt: Union[numpy.ndarray, int] = None,
        return_state: bool = None,
        blocking=True,
    ):
        """
        Vectorized version of the ``step`` method.

        It allows to step a vector of states and actions. The signature and \
        behaviour is the same as ``step``, but taking a list of states, actions \
        and dts as input.

        Args:
           actions: Iterable containing the different actions to be applied.
           states: Iterable containing the different states to be set.
           dt: int or array containing the frameskips that will be applied.
           blocking: If True, execute sequentially.
           return_state: Whether to return the state in the returned tuple. \
                If None, `step` will return the state if `state` was passed as a parameter.

        Returns:
          if states is None returns ``(observs, rewards, ends, infos)``
          else returns ``(new_states, observs, rewards, ends, infos)``

        """
        promise = self.call("step_batch", actions, states, dt, return_state)
        return promise() if blocking else promise

    def step(self, action, state=None, dt: int = 1, blocking=True):
        """
        Step the environment.

        Args:
          action: The action to apply to the environment.
          state: State to be set on the environment before stepping it.
          dt: Number of consecutive times that action will be applied.
          blocking: Whether to wait for the result.

        Returns:
          Transition tuple when blocking, otherwise callable that returns the \
          transition tuple.

        """
        promise = self.call("step", action, state, dt)
        return promise() if blocking else promise

    def reset(self, blocking=True, return_states: bool = False):
        """
        Reset the environment.

        Args:
          blocking: Whether to wait for the result.
          return_states: If true return also the initial state of the environment.

        Returns:
          New observation when blocking, otherwise callable that returns the new \
          observation.

        """
        promise = self.call("reset", return_state=return_states)
        return promise() if blocking else promise

    def _receive(self):
        """
        Wait for a message from the worker process and return its payload.

        Raises:
          Exception: An exception was raised inside the worker process.
          KeyError: The received message is of an unknown type.

        Returns:
          Payload object of the message.

        """
        message, payload = self._conn.recv()
        # Re-raise exceptions in the main process.
        if message == self._EXCEPTION:
            stacktrace = payload  # pragma: no cover
            raise Exception(stacktrace)  # pragma: no cover
        if message == self._RESULT:
            return payload
        raise KeyError("Received unexpected message {}".format(message))  # pragma: no cover

    def _worker(self, constructor, conn):
        """
        Wait for actions and send back environment results.

        Args:
          constructor: Constructor for the OpenAI Gym environment.
          conn: Connection for communication to the main process.

        Raises:
          KeyError: When receiving a message of unknown type.

        """
        try:
            env = constructor()
            env.reset()
            while True:
                try:
                    # Only block for short times to have keyboard exceptions be raised.
                    if not conn.poll(0.1):
                        continue
                    message, payload = conn.recv()
                except (EOFError, KeyboardInterrupt):  # pragma: no cover
                    break
                if message == self._ACCESS:
                    name = payload
                    result = getattr(env, name)
                    conn.send((self._RESULT, result))
                    continue
                if message == self._CALL:
                    name, args, kwargs = payload
                    result = getattr(env, name)(*args, **kwargs)
                    conn.send((self._RESULT, result))
                    continue
                if message == self._CLOSE:
                    assert payload is None
                    break  # pragma: no cover
                raise KeyError(
                    "Received message of unknown type {}".format(message),
                )  # pragma: no cover
        except Exception:  # pragma: no cover # pylint: disable=broad-except
            stacktrace = "".join(traceback.format_exception(*sys.exc_info()))
            conn.send((self._EXCEPTION, stacktrace))
            conn.close()


class BatchEnv:
    """
    Combine multiple environments to step them in batch.

    It is mostly a copy paste from \
    https://github.com/tensorflow/agents/blob/master/agents/tools/wrappers.py \
    that also allows to set and get the states.

    To step environments in parallel, environments must support a \
    ``blocking=False`` argument to their step and reset functions that \
    makes them return callables instead to receive the result at a later time.

    Args:
      envs: List of environments.
      blocking: Step environments after another rather than in parallel.

    Raises:
      ValueError: Environments have different observation or action spaces.

    """

    def __init__(self, envs, blocking):
        """
        Initialize a :class:`BatchEnv`.

        Args:
            envs: List of :class:`ExternalProcess` that contain the target environment.
            blocking: If ``True`` perform the steps sequentially. If ``False`` step \
                     the environments in parallel.

        """
        self._envs = envs
        self._blocking = blocking

    def __len__(self) -> int:
        """Return the number of combined environments."""
        return len(self._envs)

    def __getitem__(self, index):
        """Access an underlying environment by index."""
        return self._envs[index]

    def __getattr__(self, name):
        """
        Forward unimplemented attributes to one of the original environments.

        Args:
          name: Attribute that was accessed.

        Returns:
          Value behind the attribute name one of the wrapped environments.

        """
        return getattr(self._envs[0], name)

    def make_transitions(
        self,
        actions,
        states=None,
        dt: Union[numpy.ndarray, int] = 1,
        return_state: bool = None,
    ):
        """Implement the logic for stepping the environment in parallel."""
        results = []
        no_states = states is None or states[0] is None
        _return_state = ((not no_states) and return_state is None) or return_state
        chunks = ParallelEnv.batch_step_data(
            actions=actions,
            states=states,
            dt=dt,
            batch_size=len(self._envs),
        )
        for env, states_batch, actions_batch, dt in zip(self._envs, *chunks):
            result = env.step_batch(
                actions=actions_batch,
                states=states_batch,
                dt=dt,
                blocking=self._blocking,
                return_state=return_state,
            )
            results.append(result)
        results = [res if self._blocking else res() for res in results]
        return ParallelEnv.unpack_transitions(results=results, return_states=_return_state)

    def sync_states(self, state, blocking: bool = True) -> None:
        """
        Set the same state to all the environments that are inside an external process.

        Args:
            state: Target state to set on the environments.
            blocking: If ``True`` perform the update sequentially. If ``False`` step \
                     the environments in parallel.

        Returns:
            None.

        """
        for env in self._envs:
            try:
                env.set_state(state, blocking=blocking)
            except EOFError:
                continue

    def reset(self, indices=None, return_states: bool = True):
        """
        Reset the environment and return the resulting batch observations, \
        or batch of observations and states.

        Args:
          indices: The batch indices of environments to reset; defaults to all.
          return_states: return the corresponding states after reset.

        Returns:
          Batch of observations. If ``return_states`` is ``True`` return a tuple \
          containing ``(batch_of_observations, batch_of_states)``.

        """
        if indices is None:
            indices = numpy.arange(len(self._envs))
        trans = [
            self._envs[index].reset(return_states=return_states, blocking=self._blocking)
            for index in indices
        ]
        if not self._blocking:
            trans = [trans() for trans in trans]
        if return_states:
            states, obs = zip(*trans)
            states, obs = numpy.array(states), numpy.stack(obs)
            return states, obs
        return numpy.stack(trans)

    def close(self):
        """Send close messages to the external process and join them."""
        for env in self._envs:
            if hasattr(env, "close"):
                env.close()


class ParallelEnv(VectorizedEnv):
    """
    Allow any environment to be stepped in parallel when step_batch is called.

    It creates a local instance of the target environment to call all other methods.

    Example::

        >>> from plangym.videogames import AtariEnv
        >>> env = ParallelEnv(env_class=AtariEnv,
        ...                           name="MsPacman-v0",
        ...                           clone_seeds=True,
        ...                           autoreset=True,
        ...                           blocking=False)
        >>>
        >>> state, obs = env.reset()
        >>>
        >>> states = [state.copy() for _ in range(10)]
        >>> actions = [env.sample_action() for _ in range(10)]
        >>>
        >>> data =  env.step_batch(states=states, actions=actions)
        >>> new_states, observs, rewards, ends, infos = data

    """

    def __init__(
        self,
        env_class,
        name: str,
        frameskip: int = 1,
        autoreset: bool = True,
        delay_setup: bool = False,
        n_workers: int = 8,
        blocking: bool = False,
        **kwargs,
    ):
        """
        Initialize a :class:`ParallelEnv`.

        Args:
            env_class: Class of the environment to be wrapped.
            name: Name of the environment.
            frameskip: Number of times ``step`` will me called with the same action.
            autoreset: Ignored. Always set to True. Automatically reset the environment
                      when the OpenAI environment returns ``end = True``.
            delay_setup: If ``True`` do not initialize the ``gym.Environment`` \
                     and wait for ``setup`` to be called later.
            env_callable: Callable that returns an instance of the environment \
                         that will be parallelized.
            n_workers:  Number of workers that will be used to step the env.
            blocking: Step the environments synchronously.
            *args: Additional args for the environment.
            **kwargs: Additional kwargs for the environment.

        """
        self._blocking = blocking
        self._batch_env = None
        super(ParallelEnv, self).__init__(
            env_class=env_class,
            name=name,
            frameskip=frameskip,
            autoreset=autoreset,
            delay_setup=delay_setup,
            n_workers=n_workers,
            **kwargs,
        )

    @property
    def blocking(self) -> bool:
        """If True the steps are performed sequentially."""
        return self._blocking

    def setup(self):
        """Run environment initialization and create the subprocesses for stepping in parallel."""
        external_callable = self.create_env_callable(autoreset=True, delay_setup=False)
        envs = [ExternalProcess(constructor=external_callable) for _ in range(self.n_workers)]
        self._batch_env = BatchEnv(envs, blocking=self._blocking)
        # Initialize local copy last to tolerate singletons better
        super(ParallelEnv, self).setup()

    def clone(self, **kwargs) -> "PlanEnv":
        """Return a copy of the environment."""
        default_kwargs = dict(blocking=self.blocking)
        default_kwargs.update(kwargs)
        return super(ParallelEnv, self).clone(**default_kwargs)

    def make_transitions(
        self,
        actions: numpy.ndarray,
        states: numpy.ndarray = None,
        dt: Union[numpy.ndarray, int] = 1,
        return_state: bool = None,
    ):
        """
        Vectorized version of the ``step`` method.

        It allows to step a vector of states and actions. The signature and
        behaviour is the same as ``step``, but taking a list of states, actions
        and dts as input.

        Args:
            actions: Iterable containing the different actions to be applied.
            states: Iterable containing the different states to be set.
            dt: int or array containing the frameskips that will be applied.
            return_state: Whether to return the state in the returned tuple. \
                If None, `step` will return the state if `state` was passed as a parameter.

        Returns:
            if states is None returns ``(observs, rewards, ends, infos)`` else \
            ``(new_states, observs, rewards, ends, infos)``

        """
        return self._batch_env.make_transitions(
            actions=actions,
            states=states,
            dt=dt,
            return_state=return_state,
        )

    def sync_states(self, state: None):
        """
        Synchronize all the copies of the wrapped environment.

        Set all the states of the different workers of the internal :class:`BatchEnv`
         to the same state as the internal :class:`Environment` used to apply the
         non-vectorized steps.
        """
        state = self.get_state() if state is None else state
        self._batch_env.sync_states(state)

    def close(self) -> None:
        """Close the environment and the spawned processes."""
        if hasattr(self._batch_env, "close"):
            self._batch_env.close()
        if hasattr(self.gym_env, "close"):
            self.gym_env.close()



"""Module that contains environments representing video games."""
from plangym.videogames.atari import AtariEnv  # noqa: E402
from plangym.videogames.montezuma import MontezumaEnv  # noqa: E402
from plangym.videogames.nes import MarioEnv  # noqa: E402
from plangym.videogames.retro import RetroEnv  # noqa: E402



"""Implement the ``plangym`` API for Box2D environments."""
import copy

import numpy

from plangym.core import PlangymEnv


class Box2DState:
    """
    Extract state information from Box2D environments.

    This class implements basic functionalities to get the necessary
    elements to construct a Box2D state.
    """

    @staticmethod
    def get_body_attributes(body) -> dict:
        """
        Return a dictionary containing the attributes of a given body.

        Given a ``Env.world.body`` element, this method constructs a dictionary
        whose entries describe all body attributes.
        """
        base = {
            "mass": body.mass,
            "inertia": body.inertia,
            "localCenter": body.localCenter,
        }
        state_info = {
            "type": body.type,
            "bullet": body.bullet,
            "awake": body.awake,
            "sleepingAllowed": body.sleepingAllowed,
            "active": body.active,
            "fixedRotation": body.fixedRotation,
        }
        kinematics = {"transform": body.transform, "position": body.position, "angle": body.angle}
        other = {
            "worldCenter": body.worldCenter,
            "localCenter": body.localCenter,
            "linearVelocity": body.linearVelocity,
            "angularVelocity": body.angularVelocity,
        }
        base.update(kinematics)
        base.update(state_info)
        base.update(other)
        return base

    @staticmethod
    def serialize_body_attribute(value):
        """Copy one body attribute."""
        from Box2D.Box2D import b2Transform, b2Vec2

        if isinstance(value, b2Vec2):
            return tuple([*value.copy()])
        elif isinstance(value, b2Transform):
            return {
                "angle": float(value.angle),
                "position": tuple([*value.position.copy()]),
            }
        else:
            return copy.copy(value)

    @classmethod
    def serialize_body_state(cls, state_dict):
        """
        Serialize the state of the target body data.

        This method takes as argument the result given by the method
        ``self.get_body_attributes``, the latter consisting in a dictionary
        containing all attribute elements of a body. The method returns
        a dictionary whose values are the serialized attributes of the body.
        """
        return {k: cls.serialize_body_attribute(v) for k, v in state_dict.items()}

    @staticmethod
    def set_value_to_body(body, name, value):
        """Set the target value to a body attribute."""
        from Box2D.Box2D import b2Transform, b2Vec2

        body_object = getattr(body, name)
        if isinstance(body_object, b2Vec2):
            body_object.Set(*value)
        elif isinstance(body_object, b2Transform):
            body_object.angle = value["angle"]
            body_object.position.Set(*value["position"])
        else:
            setattr(body, name, value)

    @classmethod
    def set_body_state(cls, body, state):
        """
        Set the state to the target body.

        The method defines the corresponding body attribute to the value
        selected by the user.
        """
        state = state[0] if isinstance(state, numpy.ndarray) else state
        for k, v in state.items():
            cls.set_value_to_body(body, k, v)
        return body

    @classmethod
    def serialize_body(cls, body):
        """Serialize the data of the target ``Env.world.body`` instance."""
        data: dict = cls.get_body_attributes(body)
        return cls.serialize_body_state(data)

    @classmethod
    def serialize_world_state(cls, world):
        """
        Serialize the state of all the bodies in world.

        The method serializes all body elements contained within the
        given ``Env.world`` object.
        """
        return [cls.serialize_body(b) for b in world.bodies]

    @classmethod
    def set_world_state(cls, world, state):
        """
        Set the state of the world environment to the provided state.

        The method states the current environment by defining its world
        bodies' attributes.
        """
        for body, state in zip(world.bodies, state):
            cls.set_body_state(body, state)

    @classmethod
    def get_env_state(cls, env):
        """Get the serialized state of the target environment."""
        return cls.serialize_world_state(env.unwrapped.world)

    @classmethod
    def set_env_state(cls, env, state):
        """Set the serialized state to the target environment."""
        return cls.set_world_state(env.unwrapped.world, state)


class Box2DEnv(PlangymEnv):
    """Common interface for working with Box2D environments released by `gym`."""

    def get_state(self) -> numpy.array:
        """
        Recover the internal state of the simulation.

        A state must completely describe the Environment at a given moment.
        """
        state = Box2DState.get_env_state(self.gym_env)
        return numpy.array((state, None), dtype=object)

    def set_state(self, state: numpy.ndarray) -> None:
        """
        Set the internal state of the simulation.

        Args:
            state: Target state to be set in the environment.

        Returns:
            None

        """
        Box2DState.set_env_state(self.gym_env, state[0])



from typing import Tuple

import numpy
import pytest

from plangym.api_tests import batch_size, display, TestPlanEnv  # noqa: F401
from plangym.core import PlanEnv


class DummyPlanEnv(PlanEnv):
    _step_count = 0
    _state = None

    @property
    def obs_shape(self) -> Tuple[int]:
        """Tuple containing the shape of the observations returned by the Environment."""
        return (10,)

    @property
    def action_shape(self) -> Tuple[int]:
        """Tuple containing the shape of the actions applied to the Environment."""
        return tuple()

    def get_image(self):
        return numpy.zeros((10, 10, 3))

    def get_state(self):
        if self._state is None:
            state = numpy.ones(10)
            state[-1] = self._step_count
            self._state = state
            return state
        return self._state

    def set_state(self, state: numpy.ndarray) -> None:
        self._state = state

    def sample_action(self):
        return 0

    def apply_reset(self, **kwargs):
        self._step_count = 0
        return numpy.zeros(10)

    def apply_action(self, action) -> tuple:
        self._step_count += 1
        obs, reward, end, info = numpy.ones(10), 1, False, {}
        return obs, reward, end, info

    def clone(self):
        return self


environments = [lambda: DummyPlanEnv(name="dummy")]


@pytest.fixture(params=environments, scope="class")
def env(request) -> PlanEnv:
    return request.param()


@pytest.fixture(params=environments, scope="class")
def plangym_env(request) -> PlanEnv:
    return request.param()


class TestPrivateAPI:
    @pytest.mark.parametrize("dt", [1, 3])
    def test_step_with_dt(self, env, dt):
        _ = env.reset(return_state=False)
        action = env.sample_action()
        assert env.action_shape == numpy.array(action).shape
        data = env.step_with_dt(action, dt=dt)
        assert isinstance(data, tuple)



from gym.wrappers import TimeLimit
import numpy
import pytest

from plangym.environment_names import ATARI
from plangym.videogames.atari import ale_to_ram, AtariEnv
from tests import SKIP_ATARI_TESTS


if SKIP_ATARI_TESTS:
    pytest.skip("Atari not installed, skipping", allow_module_level=True)
from plangym.api_tests import (  # noqa: F401
    batch_size,
    display,
    generate_test_cases,
    TestPlanEnv,
    TestPlangymEnv,
)


def qbert_ram():
    return AtariEnv(name="Qbert-ram-v4", clone_seeds=False, autoreset=False)


@pytest.fixture(
    params=generate_test_cases(ATARI, AtariEnv, custom_tests=[qbert_ram]),
    scope="module",
)
def env(request) -> AtariEnv:
    env = request.param()
    yield env
    env.close()


class TestAtariEnv:
    def test_ale_to_ram(self, env):
        ram = ale_to_ram(env.ale)
        assert isinstance(ram, numpy.ndarray)
        assert (ram == env.get_ram()).all()

    def test_get_image(self):
        env = qbert_ram()
        obs = env.get_image()
        assert isinstance(obs, numpy.ndarray)

    def test_n_actions(self, env):
        n_actions = env.n_actions
        assert isinstance(n_actions, int)



from gym.wrappers import TimeLimit
import pytest


pytest.importorskip("Box2D")
from plangym.api_tests import (  # noqa: F401
    batch_size,
    display,
    generate_test_cases,
    TestPlanEnv,
    TestPlangymEnv,
)
from plangym.control.box_2d import Box2DEnv
from plangym.environment_names import BOX_2D


def bipedal_walker():
    timelimit = [(TimeLimit, {"max_episode_steps": 1000})]
    return Box2DEnv(name="BipedalWalker-v3", autoreset=True, wrappers=timelimit)


@pytest.fixture(
    params=generate_test_cases(BOX_2D, Box2DEnv, custom_tests=[bipedal_walker]), scope="module"
)
def env(request) -> Box2DEnv:
    return request.param()



"""Generic utilities for working with environments."""
import gym
from gym.wrappers.time_limit import TimeLimit
import numpy
from PIL import Image


def remove_time_limit_from_spec(spec):
    """Remove the maximum time limit of an environment spec."""
    if hasattr(spec, "max_episode_steps"):
        spec._max_episode_steps = spec.max_episode_steps
        spec.max_episode_steps = 1e100
    if hasattr(spec, "max_episode_time"):
        spec._max_episode_time = spec.max_episode_time
        spec.max_episode_time = 1e100


def remove_time_limit(gym_env: gym.Env) -> gym.Env:
    """Remove the maximum time limit of the provided environment."""
    if hasattr(gym_env, "spec") and gym_env.spec is not None:
        remove_time_limit_from_spec(gym_env.spec)
    if not isinstance(gym_env, gym.Wrapper):
        return gym_env
    for _ in range(5):
        try:
            if isinstance(gym_env, TimeLimit):
                return gym_env.env
            elif isinstance(gym_env.env, gym.Wrapper) and isinstance(gym_env.env, TimeLimit):
                gym_env.env = gym_env.env.env
            # This is an ugly hack to make sure that we can remove the TimeLimit even
            # if somebody is crazy enough to apply three other wrappers on top of the TimeLimit
            elif isinstance(gym_env.env.env, gym.Wrapper) and isinstance(
                gym_env.env.env,
                TimeLimit,
            ):  # pragma: no cover
                gym_env.env.env = gym_env.env.env.env
            elif isinstance(gym_env.env.env.env, gym.Wrapper) and isinstance(
                gym_env.env.env.env,
                TimeLimit,
            ):  # pragma: no cover
                gym_env.env.env.env = gym_env.env.env.env.env
            else:  # pragma: no cover
                break
        except AttributeError:
            break
    return gym_env


def process_frame(
    frame: numpy.ndarray,
    width: int = None,
    height: int = None,
    mode: str = "RGB",
) -> numpy.ndarray:
    """
    Use PIL to resize an RGB frame to a specified height and width \
    or changing it to a different mode.

    Args:
        frame: Target numpy array representing the image that will be resized.
        width: Width of the resized image.
        height: Height of the resized image.
        mode: Passed to Image.convert.

    Returns:
        The resized frame that matches the provided width and height.

    """
    height = height or frame.shape[0]
    width = width or frame.shape[1]
    frame = Image.fromarray(frame)
    frame = frame.convert(mode).resize(size=(width, height))
    return numpy.array(frame)
